package org.processmining.generalizedconformance;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.*;
import org.jgraph.JGraph;
import org.processmining.framework.util.Pair;
import org.processmining.generalizedconformance.algorithms.alignment.PrecisionAligner;
import org.processmining.generalizedconformance.algorithms.converter.ProcessTreeConverter;
import org.processmining.generalizedconformance.algorithms.converter.RelaxedPT2PetrinetConverter;
import org.processmining.generalizedconformance.algorithms.miner.QualityCriterion;
import org.processmining.generalizedconformance.algorithms.mover.SimpleScore;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.petrinets.analysis.gedsim.utils.StringEditDistance;
import org.processmining.plugins.InductiveMiner.mining.MiningParameters;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMi;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.plugins.astar.petrinet.manifestreplay.PNManifestFlattener;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.stochasticpetrinet.StochasticNetUtils;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.visualization.tree.TreeLayoutBuilder;
import org.processmining.ptconversions.pn.ProcessTree2Petrinet.PetrinetWithMarkings;
import treedist.EditScore;
import treedist.LabeledTree;
import treedist.Mapping;
import treedist.TreeEditDistance;

import java.awt.*;
import java.util.HashMap;
import java.util.Map;

public class Utils {
	
	public static final String PRECISION_MEASURE = "Precision";
	public static final String GENERALIZATION_MEASURE = "Generalization";
	
	private static final String[] nameAlphabet = new String[]{"a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","v","x","y","z","A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z","0","1","2","3","4","5","6","7","8","9"};
	

	public static ProcessTree mineProcessTree(XLog inputLog, double noiseThreshold) {
		MiningParameters parameters = new MiningParametersIMi();
		parameters.setNoiseThreshold((float)noiseThreshold);
		ProcessTree tree = IMProcessTree.mineProcessTree(inputLog, parameters);
		RelaxedPT2PetrinetConverter.postProcessMinedTree(tree);
		return tree;
	}
	
	public static XLog cloneLog(XLog log){
		XLog logClone = XFactoryRegistry.instance().currentDefault().createLog((XAttributeMap) log.getAttributes().clone());
		for (XTrace trace : log){
			XTrace traceClone = XFactoryRegistry.instance().currentDefault().createTrace((XAttributeMap) trace.getAttributes().clone());
			for (XEvent event : trace){
				XEvent eventClone = XFactoryRegistry.instance().currentDefault().createEvent((XAttributeMap) event.getAttributes().clone());
				traceClone.add(eventClone);
			}
			logClone.add(traceClone);
		}
		return logClone;
	}
	
	
	public static XLog flattenLifecycles(XLog log) {
		XLog merged = XFactoryRegistry.instance().currentDefault().createLog(log.getAttributes());
		for (XTrace trace : log){
			XTrace newTrace =  XFactoryRegistry.instance().currentDefault().createTrace(trace.getAttributes());
			for (XEvent event : trace){
				XEvent newEvent = XFactoryRegistry.instance().currentDefault().createEvent(event.getAttributes());
				String name = XConceptExtension.instance().extractName(event);
				String lc = XLifecycleExtension.instance().extractTransition(event);
				XConceptExtension.instance().assignName(newEvent, name+"_"+lc);
				newTrace.add(newEvent);
			}
			merged.add(newTrace);
		}
		return merged;
	}
	
	/**
	 * Aligns the petri net to a log and return the distance.
	 * TODO: incorporate further quality criteria!
	 * 
	 * @param petriNet
	 * @param log
	 * @return
	 */
	public static Map<QualityCriterion, Object> getDistance(PetrinetWithMarkings petriNet, XLog log) {
		double distance = 0;
		Map<QualityCriterion, Object> qualities = new HashMap<>();

		Pair<PNRepResult,PNManifestFlattener> result = (Pair<PNRepResult, PNManifestFlattener>) StochasticNetUtils.replayLog(StochasticNetUtils.getDummyConsoleProgressContext(), petriNet.petrinet, log, false, false);
		distance = Double.valueOf(result.getFirst().getInfo().get(PNRepResult.TRACEFITNESS).toString());
		qualities.put(QualityCriterion.FITNESS, distance);
		qualities.put(QualityCriterion.PRECISION1, Double.valueOf(result.getFirst().getInfo().get(PRECISION_MEASURE).toString()));
		qualities.put(QualityCriterion.GENERALIZATION1, Double.valueOf(result.getFirst().getInfo().get(GENERALIZATION_MEASURE).toString()));
		
		TransEvClassMapping evMapping = StochasticNetUtils.getEvClassMapping(petriNet.petrinet, log);
		
		PrecisionAligner precisionGen = new PrecisionAligner();
//		AlignmentPrecGenRes precGenRes = precisionGen.measureConformanceAssumingCorrectAlignment(getDummyConsoleProgressContext(), evMapping, result,
//				petriNet.petrinet, petriNet.initialMarking, false);
//		qualities.put(QualityCriterion.PRECISION, precGenRes.getPrecisionAndRecall());
//		qualities.put(QualityCriterion.GENERALIZATION, precGenRes.getGeneralization());
		qualities.put(QualityCriterion.SIMPLICITY, (double)petriNet.petrinet.getNodes().size()+petriNet.petrinet.getEdges().size());
		
//		System.out.println("Unaligned traces:");
//		// debug example misaligned trace:
//		for (SyncReplayResult repResult : result){
//			for(StepTypes stepType : repResult.getStepTypes()){
//				if (stepType.equals(StepTypes.L) || stepType.equals(StepTypes.MREAL)){
//					System.out.println(debugTrace(log.get(repResult.getTraceIndex().first())));
//					break;
//				}
//			}
//		}
		return qualities;
	}
	
	public static JGraph getGraphForTree(ProcessTree tree) {
		TreeLayoutBuilder builder = new TreeLayoutBuilder(tree);
		JGraph graph = builder.getJGraph();
		graph.setPreferredSize(new Dimension(1200,500));
		return graph;
	}
	
	public static Double getTraceSimilarity(XLog inputLog, XLog resultLog) {
		int numEventsDifferent = 0;
		int allEvents = 0;
		if (inputLog.size() != resultLog.size()){
			throw new IllegalArgumentException("Logs must have equal number of traces!");
		}
		Map<String,String> eventEncoding = new HashMap<>();
		for (int trIndex = 0 ; trIndex < inputLog.size(); trIndex++){
			XTrace trInput = inputLog.get(trIndex);
			XTrace trResult = resultLog.get(trIndex);
			String inputString = getTraceString(trInput, eventEncoding);
			String resultString = getTraceString(trResult, eventEncoding);
			allEvents += inputString.length();
			allEvents += resultString.length();
			numEventsDifferent += StringEditDistance.editDistance(inputString, resultString);
		}
		return 1 - (numEventsDifferent / (double)allEvents);
	}
	
	private static String getTraceString(XTrace trInput, Map<String, String> eventEncoding) {
		StringBuffer buf = new StringBuffer();
		for (XEvent e : trInput){
			String evName = XConceptExtension.instance().extractName(e);
			if (!eventEncoding.containsKey(evName)){
				eventEncoding.put(evName, nameAlphabet[eventEncoding.size()]);
			}
			buf.append(eventEncoding.get(evName));
		}
		return buf.toString();
	}

    /**
     * Extracts a name of an object (e.g. XLog) and returns the parameter defaultName if the name is not set
     * @param object {@link XAttributable} some object that potentially has a name
     * @param defaultName String default name
     * @return String
     */
	public static String getName(XAttributable object, String defaultName){
        return getOrDefault(XConceptExtension.instance().extractName(object), defaultName);
    }
    public static String getName(PetrinetGraph net, String defaultName){
	    return getOrDefault(net.getLabel(), defaultName);
    }
    private static String getOrDefault(String name, String defaultName) {
        if (name == null || name.trim().isEmpty()){
            name = defaultName;
        }
        return name;
    }

 	/**
 	 * Model similarity is defined as (tree edit distance / max tree edits (i.e. remove all, add all others) )  
 	 * @param inputTree {@link ProcessTree} M
 	 * @param resultTree {@link ProcessTree} M*
 	 * @return
 	 */
	public static Double getModelSimilarity(ProcessTree inputTree, ProcessTree resultTree) {
		ProcessTreeConverter converter = new ProcessTreeConverter();
		LabeledTree origLblTree = converter.getLabeledTree(inputTree);
		LabeledTree resultLblTree = converter.getLabeledTree(resultTree);
		
		Mapping map = new Mapping(origLblTree, resultLblTree);
		EditScore score = new SimpleScore(origLblTree, resultLblTree);
		TreeEditDistance dist = new TreeEditDistance(score);
		double distance = dist.calc(origLblTree, resultLblTree, map);
		return 1. - (distance / (inputTree.getNodes().size() + resultTree.getNodes().size()));
	}
}
