package org.processmining.generalizedconformance.algorithms.heuristic.footprint;


import org.processmining.generalizedconformance.algorithms.heuristic.processtree.NodeWrapper;
import org.processmining.processtree.Block;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task.Automatic;

import java.util.*;

public class ProcessTreeFootprintHandler {

	private ProcessTree tree = null;
	
	// all nodes that can generate an empty trace (eg. XOR block with one child being a tau node)
	private Set<NodeWrapper> skipNodes = new HashSet<>();

	// node -> set of nodes that can be first occurrence in subtree
	private Map<NodeWrapper, Set<NodeWrapper>> firstNodes = new HashMap<>();
	// node -> set of nodes that can be last occurrence in subtree
	private Map<NodeWrapper, Set<NodeWrapper>> lastNodes = new HashMap<>();

	public ProcessTreeFootprintHandler(ProcessTree pt) {
		this.tree = pt;
		identifySkipNodes((NodeWrapper) pt.getRoot());
		
		for (Node n : pt.getNodes()) 
			firstNodes.put((NodeWrapper) n, getFirstNodes((NodeWrapper) n));
		for (Node n : pt.getNodes()) 
			lastNodes.put((NodeWrapper) n, getLastNodes((NodeWrapper) n));
	}

	public Set<NodeWrapper> getFirstNodes(NodeWrapper n) {
		// do we know the result already?
		if (this.firstNodes.containsKey(n))
			return this.firstNodes.get(n);
		
		// no, so we need to compute it
		
		// "result" will be the nodes that can occur first in the current node
		Set<NodeWrapper> result = new HashSet<>();
		// check for trivial case
		if (n.isLeaf()) {
			result.add(n);
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.AND ||
				ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.XOR) {
			// in AND or XOR, any of the children can be first 
			for (Node child : ((Block)n).getChildren())
				result.addAll(getFirstNodes((NodeWrapper) child));	
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.LOOPXOR) {
			// in loop, the first child is first
			if (((Block)n).getChildren() != null && !((Block)n).getChildren().isEmpty()){
				NodeWrapper child = (NodeWrapper) ((Block)n).getChildren().get(0);
				result.addAll(getFirstNodes(child));
				// yet, if the first child can be skipped, the backward path can be first
				// here, we assume that there is at most one backwards path
				if (this.skipNodes.contains(child)) {
					for (int i = 1; i < ((Block)n).getChildren().size(); i++){
						Node c = ((Block)n).getChildren().get(i);
						result.addAll(getFirstNodes((NodeWrapper)c));
					}
	//				if (((Block)n).getChildren().size() > 1)
	//					result.addAll(getFirstNodes((NodeWrapper) ((Block)n).getChildren().get(1)));
				}
			}
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.SEQ) {
			// in SEQ, the first child is first
			Iterator<Node> iter = ((Block)n).getChildren().iterator();
			if (iter.hasNext()){
				NodeWrapper child = (NodeWrapper) iter.next();
				result.addAll(getFirstNodes(child));
				// yet, if the first child can be skipped, the subsequent children may also be first
				while (this.skipNodes.contains(child) && iter.hasNext()) {
					child = (NodeWrapper) iter.next();
					result.addAll(getFirstNodes(child));
				}
			}
		}
		
		// make sure to store result
		this.firstNodes.put(n, result);
		
		return result;
	}

	public Set<NodeWrapper> getLastNodes(NodeWrapper n) {
		// do we know the result already?
		if (this.lastNodes.containsKey(n))
			return this.lastNodes.get(n);
		
		// no, so we need to compute it
		
		// "result" will be the nodes that can occur last in the current node
		Set<NodeWrapper> result = new HashSet<>();
		// check for trivial case
		if (n.isLeaf()) {
			result.add(n);
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.XOR ||
				ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.AND) {
			// in AND or XOR, any of the children can be last
			for (Node child : ((Block)n).getChildren())
				result.addAll(getLastNodes((NodeWrapper) child));
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.LOOPXOR) {
			// in loop, the first child is last
			if (((Block)n).getChildren() != null && !((Block)n).getChildren().isEmpty()){
				NodeWrapper child = (NodeWrapper) ((Block)n).getChildren().get(0);
				result.addAll(getLastNodes(child));
				// yet, if the first child can be skipped, the backward path can be last
				// here, we assume that there is at most one backwards path
				if (this.skipNodes.contains(child)) {
					for (int i = 1; i < ((Block)n).getChildren().size(); i++){
						Node c = ((Block)n).getChildren().get(i);
						result.addAll(getLastNodes((NodeWrapper)c));
					}
				}
			}
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.SEQ) {
			// in SEQ, the last child is last
			List<Node> children = ((Block)n).getChildren();
			int i = children.size() - 1;
			if (i > -1){
				NodeWrapper child = (NodeWrapper) children.get(i);
				result.addAll(getFirstNodes(child));
				// yet, if the last child can be skipped, the previous children may also be last
				while (this.skipNodes.contains(child) && i > 0) {
					i--;
					child = (NodeWrapper) children.get(i);
					result.addAll(getFirstNodes(child));
				}
			}
		}
		
		// make sure to store result
		this.lastNodes.put(n, result);
		
		return result;
	}
	public void identifySkipNodes(NodeWrapper n) {
		
		// a leaf node with a tau label - record this
		if (n.isLeaf() && n.getRealNode() instanceof Automatic) {
			skipNodes.add(n);
		}
		
		// done, if n is a leaf node
		if (n.isLeaf())
			return;
		
		// n is not a leaf node, whether it can be skipped depends on the children
		for (Node child : ((Block)n).getChildren())
			identifySkipNodes((NodeWrapper) child);
		
		// decide whether n can be skipped based on its type
		if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.SEQ ||
			ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.AND) {
			// SEQ and AND require all children to occur
			// hence, they can be skipped only if all children can be skipped
			if (skipNodes.containsAll(((Block)n).getChildren()))
				skipNodes.add(n);
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.XOR) {
			// XOR require one child to occur
			// hence, it can be skipped only if at least one child can be skipped
			for (Node child : ((Block)n).getChildren()) {
				if (skipNodes.contains(child)) {
					skipNodes.add(n);
					break;
				}
			}
		}
		else if (ProcessTree.Type.getTypeOf(n.getRealNode()) == ProcessTree.Type.LOOPXOR) {
			// Loop requires the child on the forward path to occur
			// hence, it can be skipped only if the first child can be skipped
			if (((Block)n).getChildren()!=null && !((Block)n).getChildren().isEmpty()){
				if (skipNodes.contains(((Block)n).getChildren().get(0))) 
					skipNodes.add(n);
			}
		}

	}
	
	public boolean canDirectlyFollowInLoop(NodeWrapper lca, NodeWrapper t1, NodeWrapper t2) {
		// check path from ROOT to lca
		List<NodeWrapper> path = this.getDownwardPath((NodeWrapper) this.tree.getRoot(), lca);
		// check path from ROOT to parent of lca
		for (int i = 0; i < path.size() - 1; i++) {
			// for any loop that contains the lca and thus both t1 and t2 
			if (ProcessTree.Type.getTypeOf(path.get(i).getRealNode()) == ProcessTree.Type.LOOPXOR) {
				// if t1 can be last of loop and t2 can be first of loop, iteration over this loop 
				// can yield t1 being followed by t2
				// but only if the other branch can be skipped
				
				// relevant child is next in the path, except for last node in the path (in which case it is the lca)
				Node relevantChild = lca;
				if (i < path.size() - 1)
					relevantChild = path.get(i + 1);
				
				// forward or backward branch?
				if (((Block)path.get(i)).getChildren().get(0).equals(relevantChild)) {
					// forward
					
					// can we skip backward?
					if (((Block)path.get(i)).getChildren().size() < 2 || this.skipNodes.contains(((Block)path.get(i)).getChildren().get(1))) {
						// can t2 be first of forward and t1 be last of forward?
						if (this.firstNodes.get(relevantChild).contains(t2) 
								&& this.lastNodes.get(relevantChild).contains(t1))
								return true;
					}
				}
				else {
					// backward
					
					// can we skip forward?
					if (this.skipNodes.contains(((Block)path.get(i)).getChildren().get(0))) {
						// can t2 be first of backward and t1 be last of backward?
						if (this.firstNodes.get(relevantChild).contains(t2) 
								&& this.lastNodes.get(relevantChild).contains(t1))
								return true;
					}
				}
			}
		}
		return false;
	}
		
	
	/**
	 * Check if two nodes can directly follow each other in order relation
	 * @param t1 node
	 * @param t2 node
	 * @return true if t1>t2, false otherwise
	 */
	public boolean canDirectlyFollow(NodeWrapper t1, NodeWrapper t2) {
		NodeWrapper gamma = this.getLCA(t1, t2);
		
		if (gamma.isLeaf()) {
			// gamma = t1 = t2
			
			// t1 may follow itself as part of a loop over this subtree
			return canDirectlyFollowInLoop(gamma, t1, t2); 
		}
		else if (ProcessTree.Type.getTypeOf(gamma.getRealNode()) == ProcessTree.Type.XOR) {
			// cannot follow each other in this subtree
			// yet, t2 may follow t1 as part of a loop over this subtree
			return canDirectlyFollowInLoop(gamma, t1, t2); 
		}
		else if (ProcessTree.Type.getTypeOf(gamma.getRealNode()) == ProcessTree.Type.AND) {
			// t1 and t2 may occur as part of "concurrent" subtrees
			// this includes the possibility to occur directly after each other
			return true;
		}
		else if (ProcessTree.Type.getTypeOf(gamma.getRealNode()) == ProcessTree.Type.LOOPXOR) {
			// if t1 can be last of loop and t2 can be first of loop, iteration over this loop 
			// can yield t1 being followed by t2
			
			List<NodeWrapper> pathT1 = this.getDownwardPath(gamma, t1);
			List<NodeWrapper> pathT2 = this.getDownwardPath(gamma, t2);
			
			List<Node> children = ((Block)gamma).getChildren();

			Node childWithT1 = pathT1.get(1);
			Node childWithT2 = pathT2.get(1);

			// nobody forward:
			if (!children.get(0).equals(childWithT1) && !children.get(0).equals(childWithT2)){
				// both t2 and t1 backward!
				// they can only be in a follows relation, if forward can be skipped:
				return skipNodes.contains(children.get(0));
			} // t1 forward and t2 backward? 
			else if (children.get(0).equals(childWithT1)
				&& children.get(1).equals(childWithT2)) {
				if (this.lastNodes.get(childWithT1).contains(t1) 
					&& this.firstNodes.get(childWithT2).contains(t2))
					return true;
			} // t2 forward and t1 backward! 
			else {
				if (this.firstNodes.get(childWithT2).contains(t2) 
					&& this.lastNodes.get(childWithT1).contains(t1))
					return true;
			}
		}
		else if (ProcessTree.Type.getTypeOf(gamma.getRealNode()) == ProcessTree.Type.SEQ) {
			
			List<NodeWrapper> pathT1 = this.getDownwardPath(gamma, t1);
			List<NodeWrapper> pathT2 = this.getDownwardPath(gamma, t2);
			
			List<Node> children = ((Block)gamma).getChildren();
//			System.out.println(children.indexOf(pathA.get(1)));
//			System.out.println(children.indexOf(pathB.get(1)));
			
			NodeWrapper childWithT1 = pathT1.get(1);
			NodeWrapper childWithT2 = pathT2.get(1);
			int posT1 = children.indexOf(childWithT1);
			int posT2 = children.indexOf(childWithT2);
			
			// t1 comes after t2 in the sequence
			// t2 can only follow t1 directly as part of a loop over this subtree
			if (posT1 > posT2)
				return canDirectlyFollowInLoop(gamma, t1, t2);
			
			// t2 comes after t1 in the sequence
			
			// check whether they can be first/last nodes of their children
			if (this.lastNodes.get(childWithT1).contains(t1) && this.firstNodes.get(childWithT2).contains(t2)) {
				// if the child nodes are siblings, we have evidence for direct successorship
				if (posT1 == posT2 - 1)
					return true;
				// if not, we need to check whether any children in between can be skipped
				boolean canSkip = true;
				for (int i = posT1 + 1; i < posT2; i++) {
					canSkip &= this.skipNodes.contains(children.get(i));
				}
				if (canSkip)
					return true;
			}
		}
		
		return false;
	}
	
	public NodeWrapper getLCA(NodeWrapper v1, NodeWrapper v2) {
		
		if (v1.equals(v2)) return v1;
		
		List<NodeWrapper> path1 = this.getDownwardPath((NodeWrapper) this.tree.getRoot(),v1);
		List<NodeWrapper> path2 = this.getDownwardPath((NodeWrapper) this.tree.getRoot(),v2);
		
		NodeWrapper result = null;
		for (int i=0; i<path1.size(); i++) {
			if (i>=path2.size()) break;
			if (path1.get(i).equals(path2.get(i))) result = path1.get(i);
			else break;
		}
		
		return result;
	}

	public List<NodeWrapper> getDownwardPath(NodeWrapper v1, NodeWrapper v2) {		
		List<NodeWrapper> result = new ArrayList<>();
		
		NodeWrapper v = v2;
		result.add(v);
		while (!v.getIncomingEdges().isEmpty() && !result.contains(v1)) {
			v = (NodeWrapper) v.getIncomingEdges().get(0).getSource();
			result.add(v);
		}
		
		if (!result.contains(v1)) return new ArrayList<>();
		Collections.reverse(result);
		return result;
	}

}