package org.processmining.generalizedconformance.algorithms.miner;

import org.apache.commons.lang3.tuple.Pair;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.generalizedconformance.algorithms.alignment.LogAbstraction;
import org.processmining.generalizedconformance.algorithms.miner.alignment.AlignmentCollector;
import org.processmining.generalizedconformance.algorithms.miner.distance.Distance;
import org.processmining.generalizedconformance.algorithms.miner.distance.DistanceFunction;
import org.processmining.generalizedconformance.algorithms.miner.distance.SimpleDistanceFunction;
import org.processmining.generalizedconformance.algorithms.miner.move.Neighborhood;
import org.processmining.generalizedconformance.algorithms.miner.move.SearchState;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.NAryTreeImpl;
import org.processmining.plugins.stochasticpetrinet.StochasticNetUtils;

public abstract class LocalSearchMiner<M> extends OptimalMiner<M> {

    protected static final int MAX_ITER = 1000;
    protected static final int MAX_MOVES_WITHOUT_IMPROVEMENT = 20;

    protected static boolean debug = true;

    public LocalSearchMiner(DistanceFunction function, PluginContext context, XLog log, M model) {
        super(function, context, log, model);
    }

    /**
     * Simple local search meta-heuristic, where we stop searching
     * after MAX_MOVES_WITHOUT_IMPROVEMENT moves without improvement,
     * or after MAX_ITER iterations.
     * We check the neighborhood of the log and the one of the model after one another.
     */
    protected void performSearch() {
        int iter = 0;
        int movesWithoutImprovement = 0;

        Neighborhood<M> neighborhood = new Neighborhood<M>();
        visualize(currentState);
        bestState = new SearchState<>(currentState.getLogAbstraction().clone(), getClone(currentState.getModel()));
        try {
            computeDistance(currentState);
            neighborhood = recomputeNeighborhood(currentState);
        } catch (DistanceThresholdExceededException e) {
            System.err.println("This should not happen! The initial search state must be valid!");
            e.printStackTrace();
        }

        while (iter < MAX_ITER && movesWithoutImprovement < MAX_MOVES_WITHOUT_IMPROVEMENT) {
            do {
                try {
                    SearchState<M> newState = move(neighborhood, currentState);
                    visualize(newState);

                    movesWithoutImprovement++;

                    computeDistance(newState);
                    // if the previous method returns properly, it is a valid state. Only then we erase the last neighborhood and
                    // fill it anew, when we compute the log model distance


                    double newDist = function.getFinalDistance(newState.getDistance());
                    if (newDist < bestDistance){
                        System.out.println("Best distance: "+newDist);
                        movesWithoutImprovement = 0; // we improved, so we reset the counter.
                        bestDistance = newDist;
                        bestState = new SearchState<>(newState.getLogAbstraction().clone(), getClone(newState.getModel()));
                    }
                    neighborhood = recomputeNeighborhood(newState);

                    currentState = newState; // did not break, accept the new state.

                } catch (DistanceThresholdExceededException e) {
                    // move not accepted. try next one in neighborhood
                }
            } while (iter < MAX_ITER && movesWithoutImprovement < MAX_MOVES_WITHOUT_IMPROVEMENT);
        }
    }

    protected abstract Neighborhood<M> recomputeNeighborhood(SearchState<M> state);

    private void visualize(SearchState<M> state) {
        if (debug){
            visualizeState(state == null ? currentState.getModel() : state.getModel());
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    protected abstract void visualizeState(M model);

    /**
     * We assume that the alignment guides us in changing the graph and the log.
     * Needs to make sure that the traceMapping is kept consistent. If we merge a variant into another,
     * we need to reflect that.
     *
     * Also, we remove one move from the neighborhood each time we move. A move can be discarded,
     * if it leads to an infeasible state (outside of trust boundaries).
     *
     * @param neighborhood the {@link Neighborhood} consisting of possible moves (possibly with weights that reflect
     *                     their chances to improve the alignment)
     */
    protected abstract SearchState<M> move(Neighborhood<M> neighborhood, SearchState<M> currentState);
}
