package org.processmining.generalizedconformance.algorithms.miner;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.THashMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import nl.tue.astar.Trace;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.generalizedconformance.algorithms.alignment.LogAbstraction;
import org.processmining.generalizedconformance.algorithms.miner.distance.DistanceFunction;
import org.processmining.generalizedconformance.algorithms.miner.distance.DistanceHelper;
import org.processmining.generalizedconformance.algorithms.miner.move.SearchState;

import java.util.Iterator;
import java.util.Map;

public abstract class OptimalMiner<M> {

    protected final LogAbstraction origLog;
    protected final M origModel;

    protected final Map<Trace, TObjectIntMap<Trace>> traceMapping;


    protected final DistanceFunction function;

    protected double trustLog = 0.5;
    protected double trustModel = 0.5;

    protected double bestDistance = Double.MAX_VALUE;

    protected SearchState<M> bestState;

    protected SearchState<M> currentState;

    protected PluginContext context;

    protected final XEventClasses eventClasses;
    protected final Map<Integer, XEventClass> classMapping;
    protected Map<XEventClass, Integer> costMap;
    protected XLog log;


    public OptimalMiner(DistanceFunction function, PluginContext context, XLog log, M model) {
        this.log = log;
        eventClasses = getEventClasses(log);
        classMapping = new THashMap<>();

        costMap = new THashMap<>();
        int i = 0;
        for (XEventClass eClass : eventClasses.getClasses()){
            costMap.put(eClass, 2);
            classMapping.put(i++, eClass);
        }

        LogAbstraction logAbstraction = new LogAbstraction(log, eventClasses, costMap);

        this.function = function;
        this.context = context;
        this.origLog = logAbstraction.clone();
        this.origModel = getClone(model);
        this.currentState = new SearchState<>(logAbstraction, model);
        this.traceMapping = new THashMap<>();
        for (Trace trace : logAbstraction.getConvertedLog().keySet()){
            traceMapping.put(trace, new TObjectIntHashMap<Trace>());
            traceMapping.get(trace).put(trace, logAbstraction.getConvertedLog().get(trace));
        }
    }

    protected abstract M getClone(M model);

    protected static XEventClasses getEventClasses(XLog log) {
        XEventNameClassifier classifier = new XEventNameClassifier();
        XLogInfo info = XLogInfoFactory.createLogInfo(log, classifier);
        return info.getEventClasses();
    }

    /**
     * Checks the distance between log and original log and model and original model.
     * The distance is populated in the state itself.
     * Verifies that we are within our trust boundaries and throws an exception otherwise.
     * @param state the state to compute the distance for.
     * @throws DistanceThresholdExceededException when the distance is too far from the original model or log according
     *         to the given trust boundaries
     */
    protected abstract void computeDistance(SearchState<M> state) throws DistanceThresholdExceededException;

    /**
     * Levenshtein like distance
     * Assumes that traceMapping is kept consistent throughout the log edit operations!
     * @param log
     * @return
     */
    protected double getLogDistance(LogAbstraction log) {
        int totalEditOperations = 0;
        double totalEventCount = 0;
        for (Trace trace : log.getConvertedLog().keySet()){
            TObjectIntMap<Trace> map = this.traceMapping.get(trace);
            totalEventCount += log.getConvertedLog().get(trace) * trace.getSize();
            for (Trace origTrace : map.keySet()){
                int traceCount = map.get(origTrace);
                int distance = DistanceHelper.distance(trace, origTrace);
                totalEditOperations += traceCount * distance;
            }
        }
        return totalEditOperations / totalEventCount;
    }


    public XLog getBestLog() {
        return convertToLog(bestState.getLogAbstraction());
    }

    private XLog convertToLog(LogAbstraction bestLog) {
        // Decompress LogAbstraction
        Iterator<Trace> iter = bestLog.traceIterator();
        XLog resultLog = XFactoryRegistry.instance().currentDefault().createLog((XAttributeMap) log.getAttributes().clone());

        int traceCounter = 1;
        while (iter.hasNext()){
            Trace trace = iter.next();
            int instances = bestLog.getConvertedLog().get(trace);
            for (int i = 0; i < instances; i++){
                XTrace resultTrace = XFactoryRegistry.instance().currentDefault().createTrace();
                XConceptExtension.instance().assignName(resultTrace, "tr"+traceCounter++);

                addAllEvents(resultTrace, trace);
                resultLog.add(resultTrace);
            }
        }
        return resultLog;
    }

    private void addAllEvents(XTrace resultTrace, Trace trace) {
        TIntIterator iter = trace.iterator();
        while (iter.hasNext()){
            int eventType = iter.next();
            XEvent event = XFactoryRegistry.instance().currentDefault().createEvent();
            XConceptExtension.instance().assignName(event, classMapping.get(eventType).getId());
            resultTrace.add(event);
        }
    }

    public M getBestModel() {
        return bestState.getModel();
    }

    public void searchForBetterLogAndModel() {
        performSearch();
    }

    /**
     * Tries to find a log-model pair that is best in terms of a score
     */
    protected abstract void performSearch();

    public Map<Trace, TObjectIntMap<Trace>> getTraceMapping() {
        return traceMapping;
    }

    /**
     * Trust levels are the constraints on the search. If we trust our model very much,
     * we cannot move away too far from the original model in the search for a better model.
     *
     * Conversely. if we trust our log very much, we cannot correct potential noise with respect to the model.
     *
     * @param logTrust value between 0 and 1 setting the search radius (1 means no room for change - i.e., we trust the log completely)
     * @param modelTrust value between 0 and 1 setting the search radius on the model (1 means cannot move, 0 means, we do not constrain the movement)
     */
    public void setTrustLevels(double logTrust, double modelTrust) {
        this.trustLog = logTrust;
        this.trustModel = modelTrust;
    }
}
