package org.processmining.generalizedconformance.algorithms.miner;

import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.generalizedconformance.algorithms.miner.distance.DistanceFunction;
import org.processmining.generalizedconformance.algorithms.miner.move.Move;
import org.processmining.generalizedconformance.algorithms.miner.move.Neighborhood;
import org.processmining.generalizedconformance.algorithms.miner.move.SearchState;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.processtree.ProcessTree;

/**
 * Implements real local search.
 * That is, it creates a neighborhood based on the alignments and checks every neighbor for their score.
 * (Yes, by alignments!)
 *
 * Then the best neighbor is chosen in a greedy way. It is typical for this miner to get stuck in a local optimum.
 *
 * In comparison to {@link ProcessTreeLSMiner} this class computes much more alignments during moves:
 * operationCount * neighborhoodSize alignments
 *
 * The {@link ProcessTreeLSMiner} only computes operationCount alignments and choses the next neighbor randomly based
 * on the alignment-fed heuristic weights.
 *
 * Created by andreas on 5/16/17.
 */
public class ProcessTreeDumbLSMiner extends ProcessTreeLSMiner {

    public static ProcessTreeDumbLSMiner createInstance(DistanceFunction function, PluginContext context, XLog log, NAryTree model){
        ProcessTreeDumbLSMiner miner = new ProcessTreeDumbLSMiner(function, context, log, model);
        return miner;
    }

    public static ProcessTreeDumbLSMiner createInstance(DistanceFunction function, PluginContext context, XLog log, ProcessTree model){
        ProcessTreeToNAryTree treeToNAryTree = new ProcessTreeToNAryTree(getEventClasses(log));
        NAryTree nAryTree = treeToNAryTree.convert(model);
        return createInstance(function, context, log, nAryTree);
    }

    private ProcessTreeDumbLSMiner(DistanceFunction function, PluginContext context, XLog log, NAryTree model) {
        super(function, context, log, model);
    }

    @Override
    protected SearchState<NAryTree> move(Neighborhood<NAryTree> neighborhood, SearchState<NAryTree> currentState) {
        double currentDist = function.getFinalDistance(currentState.getDistance());
        // go through all neighborhoods and compute their scores. Pick best one.
        Move<NAryTree> bestMove = null;

        for (Move<NAryTree> move : neighborhood.getPossibleMoves()){
            try {
                SearchState<NAryTree> newState = move.apply(currentState, this);
                computeDistance(newState, false);

                double finalDistance = function.getFinalDistance(newState.getDistance());
                double gain = currentDist - finalDistance; // better the lower the distance gets
                move.setRealGain(gain);
                if (bestMove == null || bestMove.getRealGain() < gain){
                    bestMove = move;
                }
            } catch (DistanceThresholdExceededException e) {
                // ignore the neighbors that move outside the allowed feasible region
            }
        }

        SearchState<NAryTree> newState = currentState;
        try {
            newState = bestMove.apply(currentState, this);
        } catch (Exception e){
            System.out.println("Debug exception in ProcessTreeLSMiner.");
            e.printStackTrace();
        }

        return newState;
    }
}
