package org.processmining.generalizedconformance.algorithms.miner;

import com.mxgraph.swing.mxGraphComponent;
import gnu.trove.map.TIntIntMap;
import gnu.trove.map.TIntObjectMap;
import gnu.trove.map.hash.THashMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;
import nl.tue.astar.AStarException;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Trace;
import org.apache.commons.math3.random.MersenneTwister;
import org.apache.commons.math3.random.RandomGenerator;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.generalizedconformance.algorithms.alignment.LogAbstraction;
import org.processmining.generalizedconformance.algorithms.converter.ProcessTreeConverter;
import org.processmining.generalizedconformance.algorithms.miner.alignment.Alignment;
import org.processmining.generalizedconformance.algorithms.miner.alignment.AlignmentCollector;
import org.processmining.generalizedconformance.algorithms.miner.alignment.AlignmentPiece;
import org.processmining.generalizedconformance.algorithms.miner.alignment.NAryTreeAnnotatorReplayer;
import org.processmining.generalizedconformance.algorithms.miner.distance.Distance;
import org.processmining.generalizedconformance.algorithms.miner.distance.DistanceFunction;
import org.processmining.generalizedconformance.algorithms.miner.distance.DistanceHelper;
import org.processmining.generalizedconformance.algorithms.miner.move.*;
import org.processmining.generalizedconformance.algorithms.mover.BehaviorScore;
import org.processmining.generalizedconformance.visualizer.NAryTreeVisualization;
import org.processmining.plugins.etm.fitness.BehaviorCounter;
import org.processmining.plugins.etm.fitness.TreeFitness;
import org.processmining.plugins.etm.fitness.TreeFitnessInfo;
import org.processmining.plugins.etm.fitness.metrics.FitnessReplay;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.NAryTreeImpl;
import org.processmining.plugins.etm.model.narytree.StateBuilder;
import org.processmining.plugins.etm.model.narytree.TreeUtils;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.plugins.etm.model.narytree.replayer.AStarAlgorithm;
import org.processmining.plugins.etm.model.narytree.replayer.NAryTreeReplayer;
import org.processmining.plugins.etm.model.narytree.replayer.TreeRecord;
import org.processmining.processtree.ProcessTree;
import org.processmining.projectedrecallandprecision.framework.CompareParameters;
import treedist.LabeledTree;
import treedist.Mapping;
import treedist.TreeEditDistance;

import javax.swing.*;
import java.awt.*;
import java.util.List;
import java.util.Map;

/**
 * Process tree local search miner
 *
 * Given a process tree and a log, this miner tries to locally explore the neighborhood to find a better pair
 * of model and log. Trust values can serve as boundaries for the allowed deviations to keep close to
 * the original input.
 *
 * The optimization problem is encoded in the distance functions, which are to be minimized subject to trust constraints.
 *
 * See Rogge-Solti, Andreas, et al. "In Log and Model We Trust? A Generalized Conformance Checking Framework."
 * International Conference on Business Process Management. Springer International Publishing, 2016.
 *
 * https://link.springer.com/chapter/10.1007/978-3-319-45348-4_11
 *
 * Created by andreas on 3/22/17.
 */
public class ProcessTreeLSMiner extends LocalSearchMiner<NAryTree> {

    public static final CompareParameters parameters = new CompareParameters(2);
    private final ProcessTreeConverter converter;
    private final LabeledTree origLblTree;
    private final int logSize;
    private AStarAlgorithm logAbstraction;
    private ProcessTreeToNAryTree treeToNAryTree;

    private JFrame visualization;

    private AStarThread.Canceller canceller = new AStarThread.Canceller() {
        @Override
        public boolean isCancelled() {
            return false;
        }
    };

    private RandomGenerator randomGenerator = new MersenneTwister();

    public static final TreeFitnessInfo info = new TreeFitnessInfo(FitnessReplay.class, "Fr", "Replay Fitness",
            "Calculates the optimal alignment between the log and the process tree to determine where deviations are",
            TreeFitnessInfo.Dimension.FITNESS, true);

    private NAryTreeAnnotatorReplayer treeBasedAStar;

    public static ProcessTreeLSMiner createInstance(DistanceFunction function, PluginContext context, XLog log, NAryTree model){
        ProcessTreeLSMiner miner = new ProcessTreeLSMiner(function, context, log, model);
        return miner;
    }

    public static ProcessTreeLSMiner createInstance(DistanceFunction function, PluginContext context, XLog log, ProcessTree model){
        ProcessTreeToNAryTree treeToNAryTree = new ProcessTreeToNAryTree(getEventClasses(log));
        NAryTree nAryTree = treeToNAryTree.convert(model);
        return createInstance(function, context, log, nAryTree);
    }

    protected ProcessTreeLSMiner(DistanceFunction function, PluginContext context, XLog log, NAryTree model) {
        super(function, context, log, model);

        parameters.setDebug(false);
        this.logSize = log.size();

        converter = new ProcessTreeConverter();
        origLblTree = converter.getLabeledTree(model);

        if (debug){
            this.visualization = new JFrame("tree visualization");
            visualization.setLayout(new BorderLayout());
            visualization.setPreferredSize(new Dimension(800,800));
            visualization.setVisible(true);
            visualization.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        }
    }

    @Override
    protected NAryTree getClone(NAryTree model) {
        NAryTree clone = new NAryTreeImpl(model);
        return clone;
    }
    protected void computeDistance(SearchState<NAryTree> searchState, boolean collectAlignments) throws DistanceThresholdExceededException {
        Distance d = new Distance();
        // model to model, and log to log distance
        d.setModelDistance(getModelDistance(searchState.getModel()));
        d.setLogDistance(getLogDistance(searchState.getLogAbstraction()));
        searchState.setDistance(d);
        if (d.getLogDistance() > 1-trustLog || d.getModelDistance() > 1-trustModel){
            // reject state
            throw new DistanceThresholdExceededException("Distance threshold exceeded - rejecting state!");
        }

        computeLogModelDistance(searchState, collectAlignments);
    }

    @Override
    protected void computeDistance(SearchState<NAryTree> searchState) throws DistanceThresholdExceededException {
        computeDistance(searchState, true);
    }

    protected void computeLogModelDistance(SearchState<NAryTree> searchState, boolean collectAlignments) {
        NAryTree model = searchState.getModel();
        LogAbstraction log = searchState.getLogAbstraction();
        Distance d = searchState.getDistance();
        TreeFitness treeFitness = new TreeFitness(searchState.getModel().size());
        BehaviorCounter behC = treeFitness.behaviorCounter;

        Map<Trace, TreeRecord> alignments = null;
        if (collectAlignments) {
            alignments = new THashMap<>();
        }

        treeBasedAStar = new NAryTreeAnnotatorReplayer(log, canceller, model, -1, FitnessReplay.modelCosts(model),
                behC.getMarking2ModelMove(), behC.getMarking2VisitCount(), behC.getSyncMoveCount(), behC.getASyncMoveCount(), behC.getMoveCount(), alignments, true,
                1);

        AlignmentCollector[] alignmentCollectors = null;
        try {
            // model to log distance:
            int totalCost;
            try {
                if (treeBasedAStar != null) {
                    totalCost = treeBasedAStar.run(NAryTreeReplayer.VerboseLevel.ALL, Integer.MAX_VALUE, Integer.MAX_VALUE, -1);

                    searchState.setAlignmentCollectors(treeBasedAStar.getPartialAlignmentsPerNode());

                    d.setFitness(DistanceHelper.getFitness(model, log, behC, totalCost, eventClasses, canceller));
                    d.setPrecision(DistanceHelper.getPrecision(behC));
                }
            } catch (AStarException e) {
                e.printStackTrace();
            } catch (NullPointerException npe) {
                npe.printStackTrace();
                System.out.println("Tree: " + TreeUtils.toString(model, eventClasses));
                TreeUtils.printTree(model);
            } catch (Exception eGeneral) {
                eGeneral.printStackTrace();
                System.out.println("Tree: " + TreeUtils.toString(model, eventClasses));
                TreeUtils.printTree(model);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    protected Neighborhood<NAryTree> recomputeNeighborhood(SearchState<NAryTree> state) {
        NAryTree model = state.getModel();
        AlignmentCollector[] alignmentCollectors = state.getAlignmentCollectors();
        Neighborhood<NAryTree> neighborhood = new Neighborhood<>();
        for (int node = 0; node < model.size(); node++){
            List<Alignment> alignments = alignmentCollectors[node].getCollectedAlignments();
            if(model.isLeaf(node)){
                addLeafNeighborhood(node, alignments, model, neighborhood);
            } else {
                addConnectorNeighborhood(node, alignments, model, state.getLogAbstraction(), neighborhood);
            }
        }
        return neighborhood;
    }

    private void addConnectorNeighborhood(int node, List<Alignment> alignments, NAryTree model, LogAbstraction log, Neighborhood<NAryTree> neighborhood) {
        // go through alignments and collect evidence for insertion as choice / as parallel / at a given position
        // replay alignment on node to get the position at which a log move could be inserted in sequence

        int nChildren = model.nChildren(node);//getNumChildren(node, model);
        TIntObjectMap<int[]> logMoveMap = new TIntObjectHashMap<>();
        TIntIntMap logMoveCount = new TIntIntHashMap();

        TIntObjectMap<int[]> modelMoveMap = new TIntObjectHashMap<>();
        TIntIntMap modelMoveCount = new TIntIntHashMap();

        TIntIntMap synchMoveCount = new TIntIntHashMap();

        // collect for each log move the positions in the sequence that were enabled:
        for (Alignment alignment : alignments) {
            for (AlignmentPiece ap : alignment.getAlignment()) {
                if (ap.getModelMove()==AStarThread.NOMOVE){
                    handleMove(node, model, nChildren, logMoveMap, logMoveCount, ap.getLogMove(), ap.getMarking());
                } else if (ap.getLogMove()==AStarThread.NOMOVE) {
                    handleMove(node, model, nChildren, modelMoveMap, modelMoveCount, ap.getModelMove(), ap.getMarking());
                } else {
                    // it is unnecessary to collect the synchronous moves
                    handleMove(node, model, nChildren, null, synchMoveCount, ap.getModelMove(), ap.getMarking());
                }
            }
        }
        addNeighborhoodBasedOnPattern(node, model, logMoveMap, logMoveCount, modelMoveMap, modelMoveCount,
                synchMoveCount, alignments.size(), log, neighborhood);

        // determine position based on footprint
        // in case of sequence with one spike -> sequence) in case of multiple positions:

    }

    private void addNeighborhoodBasedOnPattern(int node, NAryTree model,
                                               TIntObjectMap<int[]> logMoveMap,
                                               TIntIntMap logMoveCount,
                                               TIntObjectMap<int[]> modelMoveMap,
                                               TIntIntMap modelMoveCount,
                                               TIntIntMap synchMoveCount,
                                               int alignmentCount,
                                               LogAbstraction log,
                                               Neighborhood<NAryTree> neighborhood) {
        int totalSyncMoves = 0;
        int totalLogMoves = 0;
        int logMovesOutsideNodes = 0;
        int logMovesExistingNodes = 0;
        TIntSet childNodes = getChildren(node, model);
        int totalModelMoves = 0;
        TIntSet childrenWithMoreThanOneOccurrence = new TIntHashSet();
        TIntSet childrenWithOneOccurrence = new TIntHashSet();
        for (int syncMoves : synchMoveCount.values()){
            totalSyncMoves += syncMoves;
        }
        for (int logMove : logMoveCount.keys()){
            int logMoves = logMoveCount.get(logMove);
            if (childNodes.contains(logMove)){
                logMovesExistingNodes += logMoves;
            } else {
                logMovesOutsideNodes += logMoves;
            }
            totalLogMoves += logMoves;
        }
        for (int modelMove : modelMoveCount.keys()){
            totalModelMoves += modelMove;
            handleChildOccurrence(log, childrenWithMoreThanOneOccurrence, childrenWithOneOccurrence, modelMove);
        }
        for (int synchMove : synchMoveCount.keys()){
            handleChildOccurrence(log, childrenWithMoreThanOneOccurrence, childrenWithOneOccurrence, synchMove);
        }

        if (logMovesOutsideNodes > 0) {
            for (int logMove : logMoveCount.keys()) {
                if (!childNodes.contains(logMove)) {
                    int positionsEnabled = 0;
                    int count = logMoveCount.get(logMove);
                    int[] logMovesPerPosition = logMoveMap.get(logMove);
                    for (int pos = 0; pos < logMovesPerPosition.length; pos++) {
                        int countAtThisPos = logMovesPerPosition[pos];
                        if (countAtThisPos > 0) {
                            positionsEnabled++;
                            if (pos == 0 || pos == model.nChildren(node)) {
                                Operation op = new Operation(Operation.INSERT, NAryTree.SEQ, pos == 0);
                                op.setNode((short) logMove);
                                neighborhood.addPossibleMove(new ModelMove(node, op, countAtThisPos, count - countAtThisPos, alignmentCount));
                            }
                        }
                    }
                    // parallel option
                    if (positionsEnabled > 1) {
                        Operation op = new Operation(Operation.INSERT, NAryTree.AND);
                        op.setNode((short) logMove);
                        neighborhood.addPossibleMove(new ModelMove(node, op, positionsEnabled, 0, logMoveMap.get(logMove).length));
                    }
                    if (logMoveCount.get(logMove) < totalModelMoves) {
                        // for the exclusive option there must be at least some model moves below as well
                        Operation op = new Operation(Operation.INSERT, NAryTree.XOR);
                        op.setNode((short) logMove);
                        neighborhood.addPossibleMove(new ModelMove(node, op, logMoveCount.get(logMove), 0, totalLogMoves));
                    }
                }
            }
        }
        short type = model.getType(node);
        if (type < 0) {
            if (type != NAryTree.LOOP) {
                // convert to loop option (equal chance to add it to the mandatory (left) / optional (right) branch of the loop)
                Operation op = new Operation(Operation.CONVERT_TO, NAryTree.LOOP, true);
                neighborhood.addPossibleMove(new ModelMove(node, op, childrenWithMoreThanOneOccurrence.size()+childrenWithOneOccurrence.size(), childrenWithOneOccurrence.size(), 0));
                Operation op2 = new Operation(Operation.CONVERT_TO, NAryTree.LOOP,false);
                neighborhood.addPossibleMove(new ModelMove(node, op2, childrenWithMoreThanOneOccurrence.size(), childrenWithOneOccurrence.size(), 0));
            }
            if (type != NAryTree.SEQ) {
                // convert to sequence option
                Operation op = new Operation(Operation.CONVERT_TO, NAryTree.SEQ);
                if (type == NAryTree.LOOP) {
                    neighborhood.addPossibleMove(new ModelMove(node, op, childrenWithOneOccurrence.size(), childrenWithMoreThanOneOccurrence.size(), 0));
                } else if (type == NAryTree.XOR || type == NAryTree.OR){
                    // many log moves of children? Then sequentializing them makes sense
                    neighborhood.addPossibleMove(new ModelMove(node, op, logMovesExistingNodes, 0, totalLogMoves+totalSyncMoves));
                } else if (type == NAryTree.AND){
                    // always allow to turn parallel into sequence ( with 0.5 weight)
                    neighborhood.addPossibleMove(new ModelMove(node, op, 1, 0, 2));
                }
            }
            if (type != NAryTree.XOR) {
                // always allow to turn connector nodes into xor
                Operation op = new Operation(Operation.CONVERT_TO, NAryTree.XOR);
                neighborhood.addPossibleMove(new ModelMove(node, op, totalModelMoves+totalLogMoves, 0, totalModelMoves+totalLogMoves+totalSyncMoves));
            }
            if (type != NAryTree.AND) {
                // always allow to turn connector nodes into parallel
                Operation op = new Operation(Operation.CONVERT_TO, NAryTree.AND);
                if (type == NAryTree.SEQ){
                    if (totalLogMoves + totalModelMoves > 0) { // otherwise we only have synchronous moves and we don't need to change a thing
                        double logMoveRatio = totalLogMoves / (totalLogMoves + totalModelMoves);
                        neighborhood.addPossibleMove(new ModelMove(node, op, 1-Math.abs(2 * logMoveRatio - 1)));
                    }
                } else if (type == NAryTree.XOR){
                    neighborhood.addPossibleMove(new ModelMove(node, op, totalModelMoves, 0, totalModelMoves+totalSyncMoves+totalLogMoves));
                } else {
                    neighborhood.addPossibleMove(new ModelMove(node, op, totalModelMoves+totalLogMoves, 0, totalModelMoves+totalLogMoves+totalSyncMoves));
                }
            }
        }
    }

    private void handleChildOccurrence(LogAbstraction log, TIntSet childrenWithMoreThanOneOccurrence,
                                       TIntSet childrenWithOneOccurrence, int activity) {
        if(log.getMaxOccurranceInTrace(activity) > 1){
            childrenWithMoreThanOneOccurrence.add(activity);
        } else {
            childrenWithOneOccurrence.add(activity);
        }
    }

    public int getState(byte[] state, int node) {
        int idx = node >>> 2;
        return (state[idx] >> 2 * (node & 3)) & 3;
    }

    /**
     * Increments counts in the data structure to capture at which position a move happened with respect to the children.
     *
     * @param node the parent node under which the non-synchronous move happens
     * @param model the current tree model
     * @param nChildren number of children of the parent node
     * @param map a counter that increments the counters at the position of an enabled child.
     *            For example, in a sequence of three elements, when there is a log move and the middle element is
     *            enabled, we count +1 at position 1 (0 is left of first child, 1 is left of second child,
     *            2 is left of third child and 3 is right of third child)
     * @param moveCount simple misalignment count per element. Can be used to select the most unaligned event or activity.
     * @param move the integer encoded event or activity that was moved asynchronously in the alignment
     * @param marking the current marking that was encountered while traversing the alignment.
     */
    private void handleMove(int node, NAryTree model, int nChildren, TIntObjectMap<int[]> map, TIntIntMap moveCount, int move, byte[] marking) {
        if (map != null  && !map.containsKey(move)){
            map.put(move, new int[nChildren+1]);
        }
        moveCount.adjustOrPutValue(move,1, 1);
        // find enabled child:
        int child = node+1;
        int childPos = 0;
        boolean found = false;
        while (model.getType(child)!=NAryTree.NONE && model.getParent(child)==node && !found) {
            if (marking != null) {
                int state = getState(marking, child);
                if (state == StateBuilder.T || state == StateBuilder.E) { // T = some element in the child is enabled, E = the child is enabled
                    found = isFound(map, move, childPos);
                }
            }
            child = model.getNext(child);
            childPos++;
        }
        if (!found && map != null){
            // move after last element in the sequence:
            map.get(move)[childPos]++;
        }
    }

    private boolean isFound(TIntObjectMap<int[]> map, int move, int childPos) {
        boolean found;
        if (map != null) {
            map.get(move)[childPos]++;
        }
        found = true;
        return found;
    }

    private TIntSet getChildren(int node, NAryTree model) {
        TIntSet children = new TIntHashSet();
        int next = model.getNext(node);
        int child = node+1;
        while (child != next){
            children.add(model.getType(child++));
        }

        return children;
    }

    /**
     * For every leaf there are the following operations.
     * 1) we can remove the leaf from the tree
     * 2) we can add another node in sequence (before or after)
     * 2) if it is a non-tau leaf, we can replace it with an optional construct
     *    (i.e., inserting a new XOR node at the position which has a TAU and the current node as children)
     * @param node
     * @param alignments
     * @param neighborhood
     */
    private void addLeafNeighborhood(int node, List<Alignment> alignments, NAryTree tree, Neighborhood<NAryTree> neighborhood) {
        int modelMoves = 0;
        int synchMoves = 0;
        TIntIntMap logMovesBefore = new TIntIntHashMap();
        TIntIntMap logMovesAfter = new TIntIntHashMap();
        TIntIntMap logMovesExclusive = new TIntIntHashMap();
        int totalLogMoves = 0;


        for (Alignment alignment : alignments){
            boolean exclusive = tree.getType(node)!=NAryTree.TAU;
            for (AlignmentPiece piece : alignment.getAlignment()) {
                if (piece.getModelMove() != AStarThread.NOMOVE && piece.getLogMove() != AStarThread.NOMOVE) {
                    exclusive = tree.getType(node)==NAryTree.TAU;
                }
            }
            boolean before = true;
            for (AlignmentPiece piece : alignment.getAlignment()) {
                if (piece.getModelMove() == AStarThread.NOMOVE) {
                    if (exclusive){
                        logMovesExclusive.adjustOrPutValue(piece.getLogMove(), alignment.getFrequency(), alignment.getFrequency());
                    } else {
                        if (before)
                            logMovesBefore.adjustOrPutValue(piece.getLogMove(), alignment.getFrequency(), alignment.getFrequency());
                        else
                            logMovesAfter.adjustOrPutValue(piece.getLogMove(), alignment.getFrequency(), alignment.getFrequency());
                    }
                    totalLogMoves += alignment.getFrequency();
                } else if (piece.getLogMove() == AStarThread.NOMOVE) {
                    modelMoves += alignment.getFrequency();
                    before = false;
                } else {
                    synchMoves += alignment.getFrequency();
                    before = false;
                }
            }
        }
        TIntSet logMoves = new TIntHashSet(logMovesBefore.keys());
        logMoves.addAll(logMovesAfter.keys());
        logMoves.addAll(logMovesExclusive.keys());
        for (int logMove : logMoves.toArray()) {
            int beforeMoves = logMovesBefore.get(logMove);
            int afterMoves = logMovesAfter.get(logMove);
            int exclusiveMoves = logMovesExclusive.get(logMove);

            if (logMove == tree.getType(node)) {
                neighborhood.addPossibleMove(new ModelMove(node, new Operation(Operation.REMOVE), modelMoves, 0, synchMoves+modelMoves));
            } else {
                Operation op = new Operation(Operation.INSERT, NAryTree.SEQ);
                op.setNode((short)logMove);
                if (beforeMoves >= afterMoves) {
                    op.setLeft(true);
                    neighborhood.addPossibleMove(new ModelMove(node, op, beforeMoves, 0, beforeMoves+afterMoves+exclusiveMoves));
                } else {
                    neighborhood.addPossibleMove(new ModelMove(node, op, afterMoves, 0, beforeMoves+afterMoves+exclusiveMoves));
                }
            }
            Operation op = new Operation(Operation.INSERT, NAryTree.XOR);
            op.setNode((short)logMove);
            neighborhood.addPossibleMove(new ModelMove(node, op, exclusiveMoves, 0, beforeMoves+afterMoves+exclusiveMoves));
        }
        if (tree.getType(node) != NAryTree.TAU && modelMoves > 0 && synchMoves > 0){
            // insert optional tau
            neighborhood.addPossibleMove(new ModelMove(node, new Operation(Operation.INSERT, NAryTree.XOR), modelMoves, synchMoves, logSize));
        } else if (alignments.isEmpty() || synchMoves == 0 && node != 0){ // do not allow to remove the root
            neighborhood.addPossibleMove(new ModelMove(node, new Operation(Operation.REMOVE), 1,0,1));
        }
    }




    /**
     * This is the alternative way
     * @param neighborhood the {@link Neighborhood} consisting of possible moves (possibly with weights that reflect
     * @param currentState
     * @return
     */
    @Override
    protected SearchState<NAryTree> move(Neighborhood<NAryTree> neighborhood, SearchState<NAryTree> currentState) {
        if (neighborhood == null || neighborhood.size() == 0){
            System.out.println("no neighborhood!");
            return currentState;
        }
        int hashCode = currentState.getModel().hashCode();
        String modelString = currentState.getModel().toString();
        double[] scores = new double[neighborhood.size()];
        int i = 0;
        double cumulativeWeight = 0;
        for (Move<NAryTree> move : neighborhood.getPossibleMoves()){
            double expectedGain = Math.max(move.getExpectedGain(),0.01);
            if (expectedGain > 0) {
                cumulativeWeight += expectedGain;
            }
            scores[i++] = expectedGain;
        }
        // randomly sample from neighborhood
        Move<NAryTree> chosenMove = null;
        double rand = randomGenerator.nextDouble()*cumulativeWeight;
        cumulativeWeight = 0;
        for (int pos = 0; pos < scores.length; pos++){
            if (scores[pos] >= 0) {
                cumulativeWeight += scores[pos];
                if (cumulativeWeight >= rand) {
                    chosenMove = neighborhood.getPossibleMoves().remove(pos);
                    break;
                }
            }
        }
        if (debug){
            System.out.println("applying " + chosenMove);
        }
        SearchState<NAryTree> newState = currentState;
        try {
            newState = chosenMove.apply(currentState, this);
        } catch (Exception e){
            System.out.println("Debug exception in ProcessTreeLSMiner.");
            e.printStackTrace();
        }
        if (!newState.getModel().toString().equals(modelString)) {
            assert newState.getModel().hashCode() != hashCode;
        }
        return newState;
    }

    @Override
    protected void visualizeState(NAryTree model) {
        mxGraphComponent graph = NAryTreeVisualization.convertToJGraph(model, eventClasses);
        graph.setPreferredSize(new Dimension(800,800));
        ScrollPane scrollPane = new ScrollPane(ScrollPane.SCROLLBARS_AS_NEEDED);
        scrollPane.add(graph);
        this.visualization.getContentPane().removeAll();
        this.visualization.getContentPane().add(scrollPane);
        this.visualization.pack();
        this.visualization.revalidate();
        this.visualization.repaint();
        return;
    }

    /**
     * Computes the distance between two models in terms of edit operations
     * @param model
     * @return
     */
    private double getModelDistance(NAryTree model) {
        if ( converter== null){
            return 0;
        }
        LabeledTree targetLblTree = converter.getLabeledTree(model);

//            if (config.getHeuristic().equals(Heuristic.BEHAVIOURAL_PROFILE)){
//                this.targetProfile = ProcessTreeBPCreator.getInstance().deriveBehaviouralProfile(targetTree);
//            } else if (config.getHeuristic().equals(Heuristic.FOOTPRINT)){
//                this.targetProfile = ProcessTreeFootprintCreator.getInstance().deriveFootprint(targetTree);
//            }

//            this.origNodeIds = getMappingOfNodes(origTree);
//            this.targetNodeIds = getMappingOfNodes(targetTree);

        Mapping map = new Mapping(origLblTree, targetLblTree);
        BehaviorScore score = new BehaviorScore(origLblTree, targetLblTree);
        TreeEditDistance dist = new TreeEditDistance(score);
        double editOps = dist.calc(origLblTree, targetLblTree, map);
        return (editOps / (Math.max(model.size(),origLblTree.size())+eventClasses.size()));
    }
}
