package org.processmining.generalizedconformance.algorithms.miner.alignment;

import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.map.TIntIntMap;
import gnu.trove.map.TIntObjectMap;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import org.processmining.plugins.boudewijn.treebasedreplay.TreeDelegate;
import org.processmining.plugins.boudewijn.treebasedreplay.astar.ModelPrefix;
import org.processmining.plugins.boudewijn.treebasedreplay.astar.TreeMarkingVisit;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.replayer.NAryTreeHead;
import org.processmining.plugins.etm.model.narytree.replayer.NAryTreePostProcessor;
import org.processmining.plugins.etm.model.narytree.replayer.NAryTreeReplayer;
import org.processmining.plugins.etm.model.narytree.replayer.TreeRecord;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Created by andreas on 4/13/17.
 */
public class AlignmentAnnotator<H extends NAryTreeHead, T extends Tail> extends NAryTreePostProcessor<H,T>{

    protected AlignmentCollector[] partialAlignmentsPerNode;

    protected TIntIntMap maxRepetitions;

    protected Map<Trace, TreeRecord> alignments;

    public AlignmentAnnotator(TreeDelegate<H, T> delegate, NAryTree tree, int configurationNumber,
                              Map<TreeMarkingVisit<ModelPrefix>, TIntSet> marking2modelmove,
                              TObjectIntMap<TreeMarkingVisit<ModelPrefix>> marking2visitCount, int[] syncMoveCount,
                              int[] aSyncMoveCount, int[] moveCount, Map<Trace, TreeRecord> alignments,
                              TIntIntMap maxRepetions) {
        super(delegate, tree, configurationNumber, marking2modelmove, marking2visitCount, syncMoveCount, aSyncMoveCount, moveCount, alignments);
        this.alignments = alignments;
        this.maxRepetitions = maxRepetions;
        partialAlignmentsPerNode = new AlignmentCollector[tree.size()];
        for (int i = 0; i < tree.size(); i++){
            partialAlignmentsPerNode[i] = new AlignmentCollector();
        }
    }

    protected void processAlignment(NAryTreeReplayer<H, T, ?> aStar, TreeRecord rec, int frequency, ModelPrefix prefix, Trace trace) {
        // update visited nodes for precision
        // and at the same time remove the base costs for moving, so only
        // the punishments remain.
        boolean[] sync = new boolean[tree.size()];
        boolean[] async = new boolean[tree.size()];

        byte[] lastMarking = findLastMarking(aStar, rec);

        TIntIntMap logMovesPerEvent = new TIntIntHashMap();
        do {
            // get the enabled moves in this state.
            TIntList enabledMoves = new TIntArrayList(10);
            // get the marking of the tree
            byte[] marking = aStar.getState(rec.getPredecessor().getState()).getHead().getState();
            lastMarking = (marking!=null && getState(marking,0)!=0) ? marking : lastMarking;
            byte[] markingToUse = marking!=null?marking:lastMarking;
            for (int node = 0; node < tree.size(); node++) {
                if (getState(lastMarking, node) == E) {
                    enabledMoves.add(node);
                    if (rec.getModelMove() == AStarThread.NOMOVE || node == rec.getModelMove()) {
                        partialAlignmentsPerNode[node].collect(new AlignmentPiece(rec.getModelMove(), getMovedEvent(trace, rec), lastMarking), frequency);
                    }
                } else if (getState(lastMarking, node) == T
                        && tree.getTypeFast(configurationNumber, node) == NAryTree.OR) {
                    enabledMoves.add(node + tree.size());
                }
                if (getState(lastMarking, node) == T) {
                    partialAlignmentsPerNode[node].collect(new AlignmentPiece(rec.getModelMove(), getMovedEvent(trace,rec), lastMarking), frequency);
                }
            }

            if (enabledMoves.isEmpty()){
                // we collect everything in the last state
                partialAlignmentsPerNode[0].collect(new AlignmentPiece(rec.getModelMove(), getMovedEvent(trace, rec), lastMarking), frequency);
            }
            if (rec.getMovedEvent()!= AStarThread.NOMOVE){
                logMovesPerEvent.adjustOrPutValue(trace.get(rec.getMovedEvent()),1,1);
            }
            // make sure to process all internal moves, also for the first element of the alignment.
            int[] internal = rec.getInternalMoves();
            for (int i = internal.length; i-- > 0;) {
                if (internal[i] >= tree.size()) {
                    // implicit OR termination, i.e. an OR terminated because all its children completed. This should
                    // be ignored in precision as the termination of this OR may not have been enabled (in case of an OR with one child)
                    //processModelMove(rec, frequency, prefix, sync, async, internal[i], enabledMoves);
                    continue;
                }
                // this node is not an OR-skip and if it is a leaf, it is hidden
                assert internal[i] < tree.size()
                        && (!tree.isLeaf(internal[i]) || tree.isHidden(configurationNumber, internal[i]));
                // check if any of the children was executed synchronously or asynchronously

                // since all children have been completed (if necessary), we can safely check the status of all
                // of them. If any of them had a synchronous move, then this node has a synchronous move and if
                // any of them has an asynchronous move, then this node has an asynchronous move.
                if (!tree.isLeaf(internal[i])) {
                    int j = internal[i] + 1;
                    do {
                        async[internal[i]] |= async[j];
                        sync[internal[i]] |= sync[j];
                        // initialize the children (for loops this may be necessary)
                        sync[j] = false;
                        async[j] = false;
                        j = tree.getNext(j);
                    } while (j < tree.getNextFast(internal[i]));
                }
                if (sync[internal[i]]) {
                    syncMoveCount[internal[i]] += frequency;
                }
                if (async[internal[i]]) {
                    aSyncMoveCount[internal[i]] += frequency;
                }
                moveCount[internal[i]] += frequency;
            }

            if (rec.getModelMove() != AStarThread.NOMOVE) {
                int m = rec.getModelMove();
                // In the state "marking", the possible moves are "enabledMoves",
                // of which "m" was executed, possibly followed by some external
                // moves stored in "rec".

                // the state is defined by the enabled moves.
                // the visit is defined in the visitCount;

                // Now check for synchronicity of the various moves.
                // We need to traverse the alignment backwards, so first process the various internal moves (if any)

                processModelMove(rec, frequency, prefix, sync, async, m, enabledMoves);

                //			System.out.println("U: " + used);
                //			System.out.println("markingVisit: " + markingVisit);

            }
            rec = rec.getPredecessor();
        } while (rec.getPredecessor() != null);

        for (int key : logMovesPerEvent.keys()){
            int maxRep = Math.max(maxRepetitions.get(key), logMovesPerEvent.get(key));
            maxRepetitions.put(key, maxRep);
        }

        for (int node = 0; node < tree.size(); node++) {
            partialAlignmentsPerNode[node].doneCollecting();
        }
    }

    private byte[] findLastMarking(NAryTreeReplayer<H, T, ?> aStar, TreeRecord rec) {
        byte[] marking = null;
        do {
            marking = aStar.getState(rec.getPredecessor().getState()).getHead().getState();
            if (marking!=null){
                // see if the root is still enabled
                int rootState = getState(marking, 0);
                if (rootState == E || rootState == T){
                    break;
                } else {
                    // continue searching for a marking where the root is enabled.
                    marking = null;
                }
            }
            rec = rec.getPredecessor();
        } while (marking == null && rec.getPredecessor()!=null);
        return marking;
    }

    private int getMovedEvent(Trace trace, TreeRecord rec) {
        if (rec.getMovedEvent() != AStarThread.NOMOVE){
            return trace.get(rec.getMovedEvent());
        }
        return rec.getMovedEvent();
    }

    @Override
    public void process(NAryTreeReplayer<H, T, ?> replayer, NAryTreeReplayer.VerboseLevel verbose, Trace trace, TreeRecord r, int frequency) {
        // Here we process an alignment to fill "marking2modelmove" and "marking2visitcount"

        if (alignments != null) {
            alignments.put(trace, r);
        }

        //  First, build the modelPrefix based on
        ModelPrefix prefix = new ModelPrefix();
        TreeRecord rec = r;
        while (rec.getPredecessor() != null) {
            if (rec.getModelMove() != AStarThread.NOMOVE && rec.getModelMove() < tree.size()
                    && tree.isLeaf(rec.getModelMove())) {
                prefix.addFirst(rec.getModelMove());
            }
            rec = rec.getPredecessor();
        }
        processAlignment(replayer, r, frequency, prefix, trace);
    }

    private void processModelMove(TreeRecord rec, int frequency, ModelPrefix prefix, boolean[] sync, boolean[] async,
                                  int m, TIntList enabledMoves) {
        if (m < tree.size()) {
            moveCount[m] += frequency;
            // we ignore the or-termination.
            if (tree.isLeaf(m)) {
                prefix.removeLast();
                // a leaf was executed.
                if (tree.getTypeFast(configurationNumber, m) == NAryTree.TAU
                        || rec.getMovedEvent() != AStarThread.NOMOVE) {
                    // this was a "skip" or a synchronous move
                    sync[m] = true;
                    syncMoveCount[m] += frequency;
                } else {
                    // this was a move on model only
                    async[m] = true;
                    aSyncMoveCount[m] += frequency;
                }
            } else {
                // a block node was executed
                // since all children have been completed (if necessary), we can safely check the status of all
                // of them. If any of them had a synchronous move, then this node has a synchronous move and if
                // any of them has an asynchronous move, then this node has an asynchronous move.
                int j = m + 1;
                do {
                    async[m] |= async[j];
                    sync[m] |= sync[j];
                    // initialize the children (for loops this may be necessary)
                    sync[j] = false;
                    async[j] = false;
                    j = tree.getNext(j);
                } while (j < tree.getNextFast(m));
                if (sync[m]) {
                    syncMoveCount[m] += frequency;
                }
                if (async[m]) {
                    aSyncMoveCount[m] += frequency;
                }
            }
        }

        TreeMarkingVisit<ModelPrefix> markingVisit = new TreeMarkingVisit<ModelPrefix>(enabledMoves, new ModelPrefix(
                prefix));

        TIntSet used = null;

        //TODO ignore internal moves that are caused by hiding nodes, these TAUs should not be counted as escaping edges!!!
        synchronized (marking2modelmove) {
            synchronized (marking2visitCount) {

                for (TreeMarkingVisit<ModelPrefix> key : marking2modelmove.keySet()) {
                    if (key.equals(markingVisit)) {
                        // state was visited before.
                        used = marking2modelmove.remove(key);
                        int visits = marking2visitCount.remove(key);
                        markingVisit.getMarking().addAll(key.getMarking());
                        marking2modelmove.put(markingVisit, used);
                        marking2visitCount.put(markingVisit, visits);
                        break;
                    }
                }
                if (used == null) {
                    used = new TIntHashSet();
                    marking2modelmove.put(markingVisit, used);
                }

				/*-
				//Log which node was used to visit a marking
				if(m < tree.size()){
					if(!marking2treenode.containsKey(markingVisit)){
						marking2treenode.put(markingVisit,new TIntHashSet());
					}
					marking2treenode.get(markingVisit).add(m);
				}/**/

                // handle the special case of m >= tree.size() indicating a skip on an OR;
                if (m >= tree.size() || tree.isLeaf(m)) {
                    marking2visitCount.adjustOrPutValue(markingVisit, frequency, frequency);
                }
                used.add(m);
            }
        }
    }

    public AlignmentCollector[] getPartialAlignmentsPerNode() {
        return partialAlignmentsPerNode;
    }
}
