package org.processmining.generalizedconformance.algorithms.miner.distance;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import gnu.trove.set.TIntSet;
import nl.tue.astar.AStarException;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Trace;
import org.deckfour.xes.classification.XEventClasses;
import org.processmining.framework.util.OsUtil;
import org.processmining.generalizedconformance.algorithms.alignment.LogAbstraction;
import org.processmining.generalizedconformance.algorithms.miner.alignment.NAryTreeAnnotatorReplayer;
import org.processmining.plugins.boudewijn.treebasedreplay.astar.ModelPrefix;
import org.processmining.plugins.boudewijn.treebasedreplay.astar.TreeMarkingVisit;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.BehaviorCounter;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.TreeUtils;
import org.processmining.plugins.etm.model.narytree.replayer.AStarAlgorithm;
import org.processmining.plugins.etm.model.narytree.replayer.NAryTreeReplayer;
import org.processmining.plugins.etm.model.narytree.replayer.TreeRecord;
import org.processmining.plugins.etm.model.narytree.replayer.basisilp2.NAryTreeReplayerWithBasis2ILP;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Created by andreas on 4/19/17.
 */
public class DistanceHelper {

    public static final double WORST_FITNESS_VALUE = 0.0;


    public static double getFScore(double recall, double precision){
        return 2 * ((recall*precision) / (recall+precision));
    }

    /**
     * Shamelessly extracted from {@link org.processmining.plugins.etm.fitness.metrics.PrecisionEscEdges}
     * @param behC
     * @return
     */
    public static double getPrecision(BehaviorCounter behC){
        // count the number of unused enabled nodes for all markings.
        double numerator = 0;
        double denominator = 0;

        synchronized (behC) {
            for (Map.Entry<TreeMarkingVisit<ModelPrefix>, TIntSet> entry : behC.getMarking2ModelMove().entrySet()) {
                int f = behC.getMarking2VisitCount().get(entry.getKey());
                int out = entry.getKey().getMarking().size();
                int used = entry.getValue().size();

                numerator += f * (double) (out - used);
                denominator += f * out;
            }
        }

        //We have no results/entries/traces aligned... this is very bad because we won't have an empty log
        if (denominator == 0) {
            return 0.0;
        }

        return 1 - (numerator / denominator);
    }

    public static double getFitness(NAryTree model, LogAbstraction logAbstraction, BehaviorCounter behC, double totalCost, XEventClasses eventClasses, AStarThread.Canceller canceller){
            // calculate the minimum model cost
            int minModelCost;

            int[] modelCosts = modelCosts(model);

            int[] syncMoveCount = new int[model.size()];
            int[] aSyncMoveCount = new int[model.size()];
            int[] moveCount = new int[model.size()];

            NAryTreeReplayer<?, ?, ?> treeBasedAStar = null;
            try {
                treeBasedAStar = setupReplayer(model, new AStarAlgorithm(eventClasses),
                        new HashMap<TreeMarkingVisit<ModelPrefix>, TIntSet>(0),
                        new TObjectIntHashMap<TreeMarkingVisit<ModelPrefix>>(0), modelCosts, syncMoveCount,
                        aSyncMoveCount, moveCount, null, canceller);
            } catch (UnsatisfiedLinkError ule) {
                System.err.println("LP Solve is not in your path so alignments cannot be calculated.");
                ule.printStackTrace();

            } catch (Exception eGeneral) {
                eGeneral.printStackTrace();
                System.out.println("Tree: " + TreeUtils.toString(model, eventClasses));
                TreeUtils.printTree(model);
                return WORST_FITNESS_VALUE;
            }
            try {
                //We don't pass a stopAt cost, but do pass a timeLimit since replaying the empty log should be easiest of all...
                minModelCost = treeBasedAStar.run(NAryTreeReplayer.VerboseLevel.NONE, Integer.MAX_VALUE, Integer.MAX_VALUE, -1);
            } catch (AStarException e) {
                e.printStackTrace();
                return WORST_FITNESS_VALUE;
            } catch (NullPointerException npe) {
                npe.printStackTrace();
                System.out.println("Tree: " + TreeUtils.toString(model, eventClasses));
                TreeUtils.printTree(model);
                return WORST_FITNESS_VALUE;
            } catch (Exception eGeneral) {
                eGeneral.printStackTrace();
                System.out.println("Tree: " + TreeUtils.toString(model, eventClasses));
                TreeUtils.printTree(model);
                return WORST_FITNESS_VALUE;
            }

            //Calculate the minimal cost for replaying the log on an 'empty' model
            int f = 0;
            double minLogCost = 0;
            Iterator<Trace> it = logAbstraction.traceIterator();
            while (it.hasNext()) {
                Trace trace = it.next();
                int freq = logAbstraction.getTraceFreq(trace);
                f += freq;
                TIntIterator it2 = trace.iterator();
                while (it2.hasNext()) {
                    minLogCost += logAbstraction.getLogMoveCost(it2.next()) * freq;
                }
            }

            //Result is the total cost divided by the 'minimal' cost without synchronous moves
            return 1.0 - (totalCost / (minLogCost + f * minModelCost));
    }

    public static int[] modelCosts(NAryTree tree) {
        int[] modelCosts = new int[tree.size()];

        int i = tree.isLeaf(0) ? 0 : tree.getNextLeafFast(0);
        do {
            if (tree.getType(i) != NAryTree.TAU) {
                //All the 'real' leafs cost something
                modelCosts[i] = 1;
            }
            i = tree.getNextLeafFast(i);
        } while (i < tree.size());
        return modelCosts;
    }

    public static NAryTreeReplayer<?, ?, ?> setupReplayer(NAryTree candidate, AStarAlgorithm algorithm,
                                                          Map<TreeMarkingVisit<ModelPrefix>, TIntSet> marking2modelMove,
                                                          TObjectIntMap<TreeMarkingVisit<ModelPrefix>> marking2visitCount, int[] modelCosts, int[] syncMoveCount,
                                                          int[] aSyncMoveCount, int[] moveCount, Map<Trace, TreeRecord> alignments, AStarThread.Canceller c) {
        NAryTreeReplayer<?, ?, ?> treeBasedAStar;
        assert candidate.isConsistent();
        try {
            treeBasedAStar = new NAryTreeReplayerWithBasis2ILP(algorithm, c, candidate, -1, modelCosts,
                    marking2modelMove, marking2visitCount, syncMoveCount, aSyncMoveCount, moveCount, alignments, true,
                    1);
			/*
			 * HV: Create a backing store that can be sufficiently large on this
			 * computer.
			 */
            long mem = OsUtil.getPhysicalMemory();
            long size = 4L * 1024 * 1024 * 1024;
            int alignment = 1;
            while (mem > size) {
                alignment++;
                mem -= size;
            }
            treeBasedAStar.setType(NAryTreeReplayer.PerformanceType.MEMORYEFFICIENT, 16 * 1024, alignment);
            treeBasedAStar.setStubborn(true);
            treeBasedAStar.setMaxNumberOfBlocksToUse(-1);
            return treeBasedAStar;
        } catch (Exception e) {
            System.out.println("Candidate: " + candidate);
            System.out.println("internal:");
            TreeUtils.printTree(candidate);
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Levenshtein distance between two int arrays
     * @param a int[] of one trace
     * @param b int[] of another trace
     * @return
     */
    public static int distance(int[] a, int[] b) {
        // i == 0
        int [] costs = new int [b.length + 1];
        for (int j = 0; j < costs.length; j++)
            costs[j] = j;
        for (int i = 1; i <= a.length; i++) {
            // j == 0; nw = lev(i - 1, j)
            costs[0] = i;
            int nw = i - 1;
            for (int j = 1; j <= b.length; j++) {
                int cj = Math.min(1 + Math.min(costs[j], costs[j - 1]), a[i - 1] == b[j - 1] ? nw : nw + 1);
                nw = costs[j];
                costs[j] = cj;
            }
        }
        return costs[b.length];
    }

    /**
     * Levenshtein distance between two Traces
     * @param a {@link Trace}
     * @param b {@link Trace}
     * @return distance in number of edit operations (count of minimally required remove / add / swap moves to get from
     * one trace to the other)
     */
    public static int distance(Trace a, Trace b) {
        // i == 0
        int [] costs = new int [b.getSize() + 1];
        for (int j = 0; j < costs.length; j++)
            costs[j] = j;
        for (int i = 1; i <= a.getSize(); i++) {
            // j == 0; nw = lev(i - 1, j)
            costs[0] = i;
            int nw = i - 1;
            for (int j = 1; j <= b.getSize(); j++) {
                int cj = Math.min(1 + Math.min(costs[j], costs[j - 1]), a.get(i - 1) == b.get(j - 1) ? nw : nw + 1);
                nw = costs[j];
                costs[j] = cj;
            }
        }
        return costs[b.getSize()];
    }
}
