package org.processmining.generalizedconformance.algorithms.miner.move;

import org.processmining.generalizedconformance.algorithms.miner.OptimalMiner;
import org.processmining.plugins.etm.model.narytree.Configuration;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.NAryTreeImpl;
import org.processmining.plugins.etm.model.narytree.TreeUtils;

/**
 * Created by andreas on 4/13/17.
 */
public class ModelMove extends Move<NAryTree> {

    private final int nodeToChange;
    private final Operation operation;

    public ModelMove(int nodeToChange, Operation op, double expectedGain){
        super(expectedGain);
        this.nodeToChange = nodeToChange;
        this.operation = op;
    }

    public ModelMove(int nodeToChange, Operation op, int affectedAlignments, int possiblyViolatedAlignments, int traceCount){
        this(nodeToChange, op, 1);
        this.expectedGain = computeExpectedGain(affectedAlignments,possiblyViolatedAlignments,traceCount);

        if (Double.isNaN(this.expectedGain)){
            throw new IllegalArgumentException("arguments NaN!");
        }
    }

    @Override
    protected double computeExpectedGain(int... arguments) {
        if (operation != null){
            if (arguments.length > 2 && arguments[2] == 0){
                arguments[2] = 1;
            }
            switch (operation.getOp()) {
                /* SUBTREE moves: */
                case Operation.CONVERT_TO:
                    switch (operation.getConnector()) {
                        case NAryTree.LOOP:
                            /* number of children that have multiple occurrences vs children that don't have multiple occurrences
                             * divided by the total number of children */
                            if (arguments[0]+arguments[1] == 0){
                                return arguments[0] - arguments[1];
                            } else {
                                return (arguments[0] - arguments[1]) / (double) (arguments[0] + arguments[1]);
                            }
                        default:
                            return (arguments[0] - arguments[1]) / (double) arguments[2];
                    }
                /* ANY moves: */
                case Operation.INSERT:
                    switch (operation.getConnector()){
                        case NAryTree.SEQ:
                        case NAryTree.AND:
                            /* log moves at different positions inside the alignment vs. positions inside the alignments */
                            // accepts m*(n+1) more words, when the current language consists of m words with average length of n.
                            // (affectedAlignments - possiblyViolatedAlignments) / (double) traceCount;
                        case NAryTree.XOR:
                        default:
                            /* Log moves vs. synchronous moves (affecting the precision) */
                            // (affectedAlignments - possiblyViolatedAlignments) / (double) traceCount;
                            return (arguments[0] - arguments[1]) / (double) arguments[2];
                    }
                /* LEAF moves: */
                case Operation.REMOVE:
                    /* Model moves vs. synchronous moves */
                    // (affectedAlignments - possiblyViolatedAlignments) / (double) traceCount;
                default:
                    /* log moves at a certain position inside the alignment vs. logmoves elsewhere of the inserted element */
                    // (affectedAlignments - possiblyViolatedAlignments) / (double) traceCount;
                    return (arguments[0] - arguments[1]) / (double) arguments[2];
            }
        }
        return 0;
    }

    @Override
    public SearchState<NAryTree> apply(SearchState<NAryTree> state, OptimalMiner<?> miner) {
        NAryTree tree = state.getModel();
        switch (operation.getOp()) {
            case Operation.REMOVE:
                tree = tree.remove(nodeToChange);
                break;
            case Operation.INSERT:
                switch (operation.getConnector()){
                    case NAryTree.XOR:
                    case NAryTree.AND:
                    case NAryTree.SEQ:
                    default:
                        NAryTree newSubTree = new NAryTreeImpl(new int[]{2, 2}, new short[]{operation.getConnector(), operation.getNode()}, new int[]{NAryTree.NONE, 0});
                        if (operation.getLeft()){
                            // the new node is left, so we add the old part on the right
                            newSubTree = newSubTree.add(tree, nodeToChange, 0, newSubTree.nChildren(0));
                        } else {
                            newSubTree = newSubTree.add(tree, nodeToChange, 0, 0);
                        }
                        tree = tree.replace(nodeToChange, newSubTree, 0);
                        break;
                }
                break;
            case Operation.CONVERT_TO:
                switch (operation.getConnector()){
                    case NAryTree.LOOP:
                        //For loops, we first need to create the loop as a tree
                        NAryTree loopTree = new NAryTreeImpl(new int[] { 4, 2, 3, 4 }, new short[] { NAryTree.LOOP,
                                NAryTree.TAU, NAryTree.TAU, NAryTree.TAU }, new int[] { NAryTree.NONE, 0, 0, 0 });
                        //Now copy the selected node into the loop body part of our loop
                        int nodeToReplace = operation.getLeft()?1:2;
                        NAryTree newLoopTree = loopTree.replace(nodeToReplace, tree, nodeToChange);
                        //And then replace the selected node by the newly created loop
                        tree = tree.replace(nodeToChange, newLoopTree, 0);
                        break;
                    default:
                        // maintain all children. They are now in sequence/XOR/AND ... :
                        tree = new NAryTreeImpl(tree);
                        tree.setType(nodeToChange, operation.getConnector());
                        break;
                }
                break;
        }
        // cleanup tree:
        tree = org.processmining.generalizedconformance.algorithms.mover.TreeUtils.cleanupTree(tree);
        assert tree.isConsistent();

        return new SearchState<>(state.getLogAbstraction(), tree);
    }

    @Override
    public String toString() {
        return "ModelMove{" +
                "node=" + nodeToChange +
                ", " + operation + '}';
    }
}
