package org.processmining.generalizedconformance.algorithms.mover;

import com.google.common.base.Joiner;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import org.jbpt.bp.RelSet;
import org.jgraph.JGraph;
import org.processmining.framework.util.Pair;
import org.processmining.generalizedconformance.algorithms.converter.ProcessTreeConverter;
import org.processmining.generalizedconformance.algorithms.heuristic.bp.BehaviourJaccardSimilarity;
import org.processmining.generalizedconformance.algorithms.heuristic.bp.ProcessTreeBPCreator;
import org.processmining.generalizedconformance.algorithms.heuristic.footprint.ProcessTreeFootprintCreator;
import org.processmining.generalizedconformance.algorithms.heuristic.processtree.NodeWrapper;
import org.processmining.generalizedconformance.algorithms.mover.EditOperation.Op;
import org.processmining.generalizedconformance.algorithms.mover.TreeMoverConfig.Heuristic;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.impl.AbstractTask;
import org.processmining.processtree.impl.AbstractTask.Manual;
import org.processmining.processtree.visualization.tree.TreeLayoutBuilder;
import treedist.LabeledTree;
import treedist.Mapping;
import treedist.TreeEditDistance;

import javax.swing.*;
import java.awt.*;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.List;

/**
 * Allows to move between two process trees along the path of a tree edit distance.
 * The edit operations are forming a spectrum between the two models.
 * 
 * @author Andreas R-Solti
 *
 */
public class TreeMover {
	
	private ProcessTreeConverter converter;
	
	private ProcessTree origTree;
	private ProcessTree targetTree;
	
	private ProcessTree resultTree;
	
	private LabeledTree origLblTree;
	private LabeledTree targetLblTree;
	
	private Map<UUID, String> names;
	
	/** maps from the orig nodes' ids to their integer position in the target labeled tree */ 
	private BiMap<UUID, Integer> origNodeIds;
	/** maps from the target nodes' ids to their integer position in the target labeled tree */ 
	private BiMap<UUID, Integer> targetNodeIds;
	
	/** maps from the original tree representation (LabeledTree) to the new nodes in the moving tree*/
	private BiMap<Integer, Node> fromOrigToNewNodes;
	/** maps from the target tree representation (LabeledTree) to the new nodes in the moving tree*/
	private BiMap<Integer, Node> fromTargetToNewNodes;
	

	private Mapping map;
	private double distance;
	private BehaviorScore score;
	
	private List<EditOperation> origEditOperations;
	
	private RelSet<ProcessTree, NodeWrapper> targetProfile;
	
	private List<EditOperation> editOperations;

	private String targetTreeString = "";

	private TreeMoverConfig config; 
	
	private boolean debugTrees = false;
	private JComponent graphPanel;
	
	public TreeMover(ProcessTree orig, ProcessTree target, TreeMoverConfig config){
		super();
		
		this.config = config;
		
		this.names = new HashMap<>();
			
		this.origTree = orig;
		this.targetTree = target;
		
		this.targetTreeString = targetTree.toString();
		
		if (debugTrees){
			this.graphPanel = new JPanel();
			JFrame frame = new JFrame("debug window");
			graphPanel.setPreferredSize(new Dimension(800,1000));
			frame.getContentPane().add(graphPanel, BorderLayout.CENTER);
			frame.pack();
			frame.setVisible(true);
			frame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		}
		
		init();
	}

	/**
	 * Precomputes the tree edit distance
	 */
	private void init() {
		this.converter = new ProcessTreeConverter();
		
		this.origLblTree = converter.getLabeledTree(origTree);
		this.targetLblTree = converter.getLabeledTree(targetTree);
		
		if (config.getHeuristic().equals(Heuristic.BEHAVIOURAL_PROFILE)){
			this.targetProfile = ProcessTreeBPCreator.getInstance().deriveBehaviouralProfile(targetTree);
		} else if (config.getHeuristic().equals(Heuristic.FOOTPRINT)){
			this.targetProfile = ProcessTreeFootprintCreator.getInstance().deriveFootprint(targetTree);
		}
		
		this.origNodeIds = getMappingOfNodes(origTree);
		this.targetNodeIds = getMappingOfNodes(targetTree);
		
		this.map = new Mapping(this.origLblTree, this.targetLblTree);
		this.score = new BehaviorScore(this.origLblTree, this.targetLblTree);
		TreeEditDistance dist = new TreeEditDistance(this.score);
		distance = dist.calc(this.origLblTree, this.targetLblTree, map);
		
		this.origEditOperations = getEditOperations();
		if (config.getHeuristic().equals(Heuristic.RANDOM)){
			Collections.shuffle(origEditOperations);
		}
//		TreeDistanceTest.debugTrace(converter, origLblTree, targetLblTree, map);
	}

	private BiMap<UUID,Integer> getMappingOfNodes(ProcessTree tree) {
		BiMap<UUID,Integer> mapping = HashBiMap.create();
		for (Node node : tree.getNodes()){
			String nodeName = node.getName();
			nodeName = nodeName.trim().isEmpty()? node.getClass().getSimpleName(): nodeName;
			mapping.put(node.getID(), converter.getNodeMapping().get(node));
			names.put(node.getID(), nodeName);
		}
		return mapping;
	}
	
	public ProcessTree getProcessTreeBasedOnTrust(double trustLevelLog, double trustLevelModel){
		double distanceRatio = getTrustLevelDistance(trustLevelModel, trustLevelLog);
		
		if (distanceRatio < 0 || distanceRatio > 1){
			throw new IllegalArgumentException("trust level must be between 1 (total trust) and 0 (no trust at all)");
		}
		
		this.resultTree = TreeUtils.getClone(this.origTree);
		
		this.fromOrigToNewNodes = HashBiMap.create();
		this.fromTargetToNewNodes = HashBiMap.create();
		for (Node n : resultTree.getNodes()){
			fromOrigToNewNodes.put(converter.getNodeMapping().get(n), n);
		}
		for (int[] repl : map.getAllReplacement()){
			if (fromOrigToNewNodes.containsKey(repl[0])){
				fromTargetToNewNodes.put(repl[1], fromOrigToNewNodes.get(repl[0]));
				System.out.println("mapped: "+names.get(origNodeIds.inverse().get(repl[0]))+" - "+names.get(targetNodeIds.inverse().get(repl[1])));
			} else {
				System.out.println("Debug this! - all original replacement nodes should be already in the map.");
			}
		}
		
		// add edit operations from the mapping, until we reach a point close to the trust level
		editOperations = getEditOperations();

		if (debugTrees){
			visualizeTree(this.resultTree, this.targetTree);
		}

		double targetDistance = this.distance * distanceRatio;
				
		double currentDistance = 0;
		
		if (config.getHeuristic().equals(Heuristic.SIMPLE) || config.getHeuristic().equals(Heuristic.RANDOM)){
			ProcessTree previousTree = TreeUtils.getClone(this.resultTree);
			while (!editOperations.isEmpty() && currentDistance < targetDistance){
				for (EditOperation oper : editOperations){
					//if (isAllowedOp(oper, editOperations, resultTree)){
						currentDistance += apply(oper, resultTree);
						editOperations.remove(oper);
						if (debugTrees)
							visualizeTree(previousTree, resultTree, targetTree);
						previousTree = TreeUtils.getClone(this.resultTree);
						break;
					//}
				}
			}
		} else {
			ProcessTree previousTree = TreeUtils.getClone(this.resultTree);
			while (!editOperations.isEmpty() && (trustLevelModel == 0 || currentDistance < targetDistance)){
				// check all possible next moves and apply the best one
				Pair<EditOperation, Double> bestEditOperation = null;
				for (EditOperation oper : editOperations){
					try {
						if (isAllowedOp(oper, editOperations, resultTree)){
							// peek forward to see, how good the next tree would be.
							List<Edge> edges = new ArrayList<>(resultTree.getEdges());
							Collections.sort(edges, new Comparator<Edge>() {
								public int compare(Edge o1, Edge o2) {
									return o1.toString().compareTo(o2.toString());
								}
							});
							int size = edges.size();
							int nodeSize = resultTree.getNodes().size();
							String treeString = resultTree.toString();
							
							apply(oper, resultTree);
							if (!checkTreeConsistent(resultTree) || !validateLeafs(resultTree)){
								System.out.println("Operation "+oper.toString()+" broke tree!");
								System.out.println(resultTree);
							}
							String intermediate = resultTree.toString();
							RelSet<ProcessTree, NodeWrapper> potentialProfile = null;
							if (config.getHeuristic().equals(Heuristic.BEHAVIOURAL_PROFILE)) {
								potentialProfile = ProcessTreeBPCreator.getInstance().deriveBehaviouralProfile(resultTree);
							} else if (config.getHeuristic().equals(Heuristic.FOOTPRINT)){
								potentialProfile = ProcessTreeFootprintCreator.getInstance().deriveFootprint(resultTree);
							}
							unapply(oper, resultTree);
							reset(oper);
							if (!checkTreeConsistent(resultTree) || !validateLeafs(resultTree) || !resultTree.toString().equals(treeString)){
								System.err.println("Operation "+oper.toString()+" broke tree!");
								System.out.println("orig: "+treeString);
								System.out.println("then: "+intermediate);
								System.out.println("new : "+resultTree.toString());
								apply(oper, resultTree);
								unapply(oper, resultTree);
							}
							List<Edge> newEdges = new ArrayList<>(resultTree.getEdges());
							Collections.sort(newEdges, new Comparator<Edge>() {
								public int compare(Edge o1, Edge o2) {
									return o1.toString().compareTo(o2.toString());
								}
							});
							if (resultTree.getEdges().size()!=size || resultTree.getNodes().size() != nodeSize){
								System.out.println("before oper "+oper+" : edges("+edges.size()+"), nodes("+nodeSize+")");
								System.out.println("after oper "+oper+" : edges("+resultTree.getEdges().size()+"), nodes("+resultTree.getNodes().size()+")");
								System.out.println(Joiner.on("\n ").join(edges));
								System.out.println("");
								System.out.println(Joiner.on("\n ").join(newEdges));
							}
							
							Pair<EditOperation, Double> thisOperation = new Pair<>(oper, getSimilarityOfProfiles(targetProfile, potentialProfile));
							if (bestEditOperation == null || bestEditOperation.getSecond() < thisOperation.getSecond()){
								bestEditOperation = thisOperation;
							}
						}
					} catch (IllegalArgumentException e){
						e.printStackTrace();
						// ignore moves that violate the behavioral profile computation
						// for example: adding an already existing node, before removing the old one.
					}
				}
				if (bestEditOperation == null){
					EditOperation toBreak = editOperations.remove(0);
					
					// we face a permutation renaming. Need to break the cycle by converting a mapping x<->y to x->0, 0->y and apply x->0.
					EditOperation deleteOp = new EditOperation(Op.DELETE, toBreak.getOrigNode(), toBreak.getOrigName(), null, null, score.delete(origNodeIds.get(toBreak.getOrigNode())));
					EditOperation insertOp = new EditOperation(Op.INSERT, null, null, toBreak.getNewNode(), toBreak.getNewName(), score.insert(targetNodeIds.get(toBreak.getNewNode())));
					editOperations.add(deleteOp);
					editOperations.add(insertOp);
				} else {
					currentDistance += apply(bestEditOperation.getFirst(), resultTree);
					if (!targetTreeString.equals(targetTree.toString())){
						System.err.println("Target Tree broken!!");
						System.out.println(bestEditOperation.getFirst());
					}
					if (debugTrees)
						visualizeTree(previousTree, resultTree, targetTree);
					editOperations.remove(bestEditOperation.getFirst());
					previousTree = TreeUtils.getClone(resultTree);
				}
			}	
		}
		
		return resultTree;
	}


	private void reset(EditOperation oper) {
		oper.setChildIndices(null);
		oper.setParentIndex(null);
		oper.setParentNode(null);
	}

	private boolean checkTreeConsistent(ProcessTree resultTree2) {
		boolean consistentParents = true;
		for (Node node : resultTree2.getNodes()) {
			if (resultTree2.getRoot().equals(node)) {
				if (node.getParents().size() != 0){
					consistentParents = false;
					System.out.println("Debug me!");	
				}
			} else if (node.getParents().size() != 1) {
				consistentParents = false;
				System.out.println("Debug me!");
			}
		}
		return consistentParents;
	}

	/**
	 * In case a node should be relocated by two add / remove operations, we only allow to do remove first!
	 * Same goes for rename operations to  
	 * @param oper
	 * @param editOperations
	 * @param resultTree2
	 * @return
	 */
	private boolean isAllowedOp(EditOperation oper, List<EditOperation> editOperations, ProcessTree resultTree2) {
		String nameToCheck = null;
		if (oper.getOperation().equals(Op.INSERT) || oper.getOperation().equals(Op.RENAME)){
			nameToCheck = oper.getNewName();
		}
		if (getLeafNames(resultTree2).contains(nameToCheck)){
			return false;
		}
		if (oper.getOperation().equals(Op.DELETE)){
			Integer origId = origNodeIds.get(oper.getOrigNode());
			Node origNode = fromOrigToNewNodes.get(origId);
			if (resultTree2.getRoot().equals(origNode)){
				if (origNode instanceof Block){
					// only allow to remove roots that have a single child
					Block origBlock = (Block) origNode;
					return origBlock.getChildren().size() == 1;	
				} else {
					// root node to remove is the only leaf in the tree.
					return true;
				}
			}
		}
		return true;
		
	}

	private Double getSimilarityOfProfiles(RelSet<ProcessTree, NodeWrapper> profileA,
			RelSet<ProcessTree, NodeWrapper> profileB) {
		BehaviourJaccardSimilarity<ProcessTree, NodeWrapper> sim = new BehaviourJaccardSimilarity<>(profileA, profileB);
		return sim.getSimilarity();
//		return NameBasedSimilarity.getNameBasedSimilarity(profileA, profileB);
	}
	
	public double unapply(EditOperation op, ProcessTree tree){
		op.setReverse(!op.isReverse());
		double val = apply(op, tree);
		op.setReverse(!op.isReverse());
		return val;
	}

	public double apply(EditOperation op, ProcessTree tree){
		System.out.println("Applying: "+op);
//		visualizeTree(tree, targetTree);
		int nodeCount = tree.getNodes().size();
		int edgeCount = tree.getEdges().size();
		switch (op.getOperation()){
			case INSERT:
				if (!op.isReverse()){
					insertNodeIntoTree(op.getNewNode(), false, tree, op);
					checkTreeModification(tree, nodeCount, edgeCount, op);
				} else {
					// reverse insert = delete
					deleteNodeFromTree(op.getNewNode(), true, tree, op);
					checkTreeModification(tree, nodeCount, edgeCount, op);
				}
				break;
			case DELETE:
				if (!op.isReverse()){
					deleteNodeFromTree(op.getOrigNode(), false, tree, op);
					checkTreeModification(tree, nodeCount, edgeCount, op);
				} else {
					insertNodeIntoTree(op.getOrigNode(), true, tree, op);
					checkTreeModification(tree, nodeCount, edgeCount, op);
				}
				break;
			case RENAME:
				if (!op.isReverse()){
					renameNodesInTree(op.getOrigNode(), op.getNewNode(), false, tree);
					checkTreeModification(tree, nodeCount, edgeCount, op);
				} else {
					renameNodesInTree(op.getNewNode(), op.getOrigNode(), true, tree);
					checkTreeModification(tree, nodeCount, edgeCount, op);
				}
				break;
			default:
				break;
				
		}
//		visualizeTree(tree, targetTree);
		return op.getCost();
	}

	private boolean validateLeafs(ProcessTree tree) {
		
		List<String> leafs = new ArrayList<>();
		for (Node node : tree.getNodes()){
			if (node instanceof Manual){
				String name = node.getName(); 
				if (leafs.contains(name)){
					return false;
				} else {
					leafs.add(name);
				}
			}
		}
		return true;
	}

	private Set<String> getLeafNames(ProcessTree tree){
		Set<String> names = new HashSet<>();
		for (Node node : tree.getNodes()){
			if (node instanceof Manual){
				names.add(node.getName());
			}
		}
		return names;
	}
	
	private boolean checkTreeModification(ProcessTree tree, int nodeCount, int edgeCount, EditOperation operation) {
		int expectedChange = operation.getOperation().equals(Op.DELETE) ? -1 : ( operation.getOperation().equals(Op.INSERT) ? 1 : 0);
		if (operation.isReverse()) expectedChange *= -1;
		if (tree.getNodes().size() != nodeCount+expectedChange || tree.getEdges().size() != edgeCount + expectedChange){
			System.out.println(operation+ " broke tree size constraints!");
			System.out.println(tree);
			return false;
		}
		return true;
	}

	/**
	 * This corresponds to a replacement mapping. 
	 * We find the original node in the tree, and replace it with the new node.
	 * 
	 * All connections are passed to the new node.
	 * 
	 * @param origNodeId
	 * @param newNodeId
	 * @param reverse
	 * @param tree
	 */
	protected void renameNodesInTree(UUID origNodeId, UUID newNodeId, boolean reverse, ProcessTree tree) {
		BiMap<UUID, Integer> fromNodeIds = reverse?targetNodeIds:origNodeIds;
		BiMap<UUID, Integer> toNodeIds = reverse?origNodeIds:targetNodeIds;
		BiMap<Integer, Node> fromNodeToNewNodes = reverse?fromTargetToNewNodes:fromOrigToNewNodes;
		BiMap<Integer, Node> toNodeToNewNodes = reverse?fromOrigToNewNodes:fromTargetToNewNodes;
		
		
		int nodeIdToRename = fromNodeIds.get(origNodeId);
		int nodeIdToRenameTo = toNodeIds.get(newNodeId);

		Node nodeToRename = fromNodeToNewNodes.get(nodeIdToRename);
		int nodePos = nodeToRename.getParents().isEmpty()? 0 : getPositionInParent(nodeToRename, nodeToRename.getParents().iterator().next()); 
		Node newNode = reverse ? origTree.getNode(newNodeId) : targetTree.getNode(newNodeId);

		
		Node np = getNewNode(newNode, nodeToRename, tree);
		
		if (np != null && nodeToRename != null && np instanceof Block && nodeToRename instanceof Block){
			// transfer all connection of the original node to the new node:
			Block origBlock = (Block) nodeToRename;
			Block newBlock = (Block) np;
			
			List<Edge> edgesToRelocate = new ArrayList<>(origBlock.getOutgoingEdges());
			for (Edge e : edgesToRelocate){
				tree.removeEdge(e);
				origBlock.removeOutgoingEdge(e);
				e.getTarget().removeIncomingEdge(e);
				newBlock.addChild(e.getTarget());
			}
		}
		if (!nodeToRename.getParents().isEmpty()){
			Block parent = nodeToRename.getParents().iterator().next();
			parent.addChildAt(np, nodePos);
			Collection<Edge> edgesToRemove = new ArrayList<>();
			for (Edge outgoing : parent.getOutgoingEdges()){
				if (outgoing.getTarget().equals(nodeToRename)){
					edgesToRemove.add(outgoing);
				}
			}
			for (Edge toRemove : edgesToRemove){
				parent.removeOutgoingEdge(toRemove);
				toRemove.getTarget().removeIncomingEdge(toRemove);
				tree.removeEdge(toRemove);
			}
		}
		
		if (nodeToRename.getParents().size()>0){
			System.out.println("Debug me!");
		}
		
		tree.removeNode(nodeToRename);

		if (tree.getRoot().equals(nodeToRename)){
			tree.setRoot(np);
		}
		
		if (np.getParents().size()>1){
			System.out.println("Debug me!");
		}
		
//			toRemove.setProcessTree(null);
		
		fromNodeToNewNodes.put(nodeIdToRename, np);
		toNodeToNewNodes.put(nodeIdToRenameTo, np);
	}

	/**
	 * For deleting a node, we need to:
	 * - know the parent of the node to be deleted.
	 * - know the children of the node to be deleted.
	 * - delete the node and connect the children to the parent of the deleted node (in the correct position!)
	 * 
	 * @param uuid
	 * @param reverse
	 * @param tree
	 * @return int index in parent node
	 */
	protected void deleteNodeFromTree(UUID uuid, boolean reverse, ProcessTree tree, EditOperation op) {
		BiMap<UUID, Integer> nodeIds = reverse ? targetNodeIds : origNodeIds ;
		BiMap<Integer, Node> fromNodeToNewNodes = reverse?fromTargetToNewNodes:fromOrigToNewNodes;
		
		Integer nodeIdInOrig = nodeIds.get(uuid);
		Node origNode = fromNodeToNewNodes.get(nodeIdInOrig);
		
		if (origNode == null){
			System.out.println("Debug me!");
		}
		if (tree.getRoot().equals(origNode) && origNode instanceof Block){
			// check whether we have more than 1 children - then we can't perform the removal.
			Block origBlock = (Block) origNode;
			if (origBlock.getOutgoingEdges().size()==1){
				Edge edgeToChild = origBlock.getOutgoingEdges().get(0);
				Node childThatWillBecomeRoot = edgeToChild.getTarget();
				origBlock.removeOutgoingEdge(edgeToChild);
				childThatWillBecomeRoot.removeIncomingEdge(edgeToChild);
				tree.removeEdge(edgeToChild);
				tree.removeNode(origBlock);
				tree.setRoot(childThatWillBecomeRoot);
				op.setParentIndex(-1);
			} else {
				throw new IllegalArgumentException("Can't delete root node, as long as it has more than one child!");
			}
		} else {
			Node parentNode = origNode.getParents().iterator().next();
			op.setParentNode(parentNode);
			int currentSiblingPosition = getPositionInParent(origNode, parentNode);
			int parentIndex = currentSiblingPosition;
			
			List<Edge> edgesToRelocate = new ArrayList<>();
			// we need to find all children that are below the old node and are currently present in the tree
			if (origNode instanceof Block){ // might have children
				Block origBlock = (Block) origNode;
				edgesToRelocate = new ArrayList<>(origBlock.getOutgoingEdges());
				int i = 0;
				if (parentNode != null){
					Block newParentBlock = (Block) parentNode;
					for (Edge e : edgesToRelocate){
						currentSiblingPosition = (op.getChildIndices()==null)?currentSiblingPosition : op.getChildIndices()[i++];
						if (reverse && op.getParentIndex()!=null && op.getParentIndex()<currentSiblingPosition){
							currentSiblingPosition++; // add one for nodes right of the current node
						}
						newParentBlock.addChildAt(e.getTarget(), currentSiblingPosition++);
						e.getTarget().removeIncomingEdge(e);
						e.getSource().removeOutgoingEdge(e);
						tree.removeEdge(e);
					}
				}
			}
			if (origNode.getParents().size()>1){
				System.out.println("Debug me!");
			}
			
			// add children at sibling position of current node that will be removed
			if (parentNode != null){
				Block newParentBlock = (Block) parentNode;
				
				Edge toRemove = null;
				for (Edge e : newParentBlock.getOutgoingEdges()){
					if (e.getTarget().getID().equals(origNode.getID())){
						toRemove = e;
					}
				}
				newParentBlock.removeOutgoingEdge(toRemove);
				tree.removeEdge(toRemove);
				tree.removeNode(origNode);
				fromNodeToNewNodes.remove(nodeIdInOrig);
			} else {
				System.out.println("Debug me!");
			}
			op.setParentIndex(parentIndex);
		}
	}

	/**
	 * When inserting a node, it is important to find a mapped parent of it.
	 * We need to insert the node under that parent.
	 * Then, we check all children of that parent, whether they should be descendants of the new node.
	 * These will be put below the new node and their edges to the parent node are cut.
	 *  
	 * @param uuid
	 * @param reverse
	 * @param tree
	 */
	protected void insertNodeIntoTree(UUID uuid, boolean reverse, ProcessTree tree, EditOperation oper) {
		Integer parentIndex = oper.getParentIndex();
		BiMap<UUID, Integer> nodeIds = reverse ? origNodeIds : targetNodeIds;
		BiMap<Integer, Node> nodeToNewNodes = reverse?fromOrigToNewNodes:fromTargetToNewNodes;
		ProcessTree sourceTree = reverse? origTree : targetTree; 
		
		if (reverse && parentIndex == -1){
			// we reinsert a former root node!
			Integer nodeIdInTree = nodeIds.get(uuid);
			Node oldRoot = tree.getRoot();
			Node newRoot = sourceTree.getNode(uuid);
			Node newRootToInsert = getNewNode(newRoot, null, tree);
			tree.addNode(newRootToInsert);
			Block newBlock = (Block) newRootToInsert;
			Edge childEdge = newBlock.addChild(oldRoot);
			oldRoot.addIncomingEdge(childEdge);
			tree.setRoot(newRootToInsert);
			nodeToNewNodes.remove(nodeIdInTree);
			nodeToNewNodes.put(nodeIdInTree, newRootToInsert);
		} else {
		
			Integer nodeIdInTree = nodeIds.get(uuid);
			Node newNode = sourceTree.getNode(uuid);
	//		if (newNode instanceof Automatic && !reverse){
	//			System.out.println("debug here");
	//		}
			
			LabeledTree labeledTree = reverse ? origLblTree : targetLblTree;
			
			// find next existing parent under which we can place the new node:
			int newParentInTree = findNextMappedParent(nodeIdInTree, labeledTree, nodeToNewNodes);
			Node parentNode = nodeToNewNodes.get(newParentInTree);
			if (reverse && oper.getParentNode() != null){
				parentNode = oper.getParentNode();
			}
			List<Integer> edgeIndicesToRelocate = new ArrayList<>();
			// we need to find all children that will go below the new node and are currently present in the tree
			// it might be that some children are hidden below other nodes that will be removed
			// furthermore, the children could be still subject to a renaming
			if (parentNode != null && parentNode instanceof Block){ // has parent
				Block parentBlock = (Block) parentNode;
				
				int lastChild = 0;
				List<Edge> edges = parentBlock.getOutgoingEdges();
				List<Edge> edgesToRelocate = new ArrayList<>();
				int index = 0;
				for (Edge outgoing : edges){
					// TODO: need to check all children!
					Node child = outgoing.getTarget();
					Integer childIdInTarget =  nodeToNewNodes.inverse().get(child);
					boolean childContainsDescendants = containsDescendants(child,nodeIdInTree, labeledTree, nodeToNewNodes); 
					if (childIdInTarget == null){
						// go through all the node's children recursively
						if (childContainsDescendants){
							lastChild = Math.max(lastChild, getPositionInParent(child, parentBlock))+1;
							edgesToRelocate.add(outgoing);
							edgeIndicesToRelocate.add(index);
						}
						
						// not added yet, or belongs to other tree 
						BiMap<Integer, Node> nodeToOldNodes = reverse?fromTargetToNewNodes:fromOrigToNewNodes;
						Integer childInSource = nodeToOldNodes.inverse().get(child);
						if (childInSource == null){
							System.out.print(".");
							// not added yet!
						} else {
							if (isRightOf(childInSource, nodeIdInTree, reverse ? targetLblTree:origLblTree)){
								lastChild++;
							}
						}
					} else {
						if (isAncestor(childIdInTarget, nodeIdInTree, labeledTree)){
							lastChild = Math.max(lastChild, getPositionInParent(child, parentBlock))+1;
							edgesToRelocate.add(outgoing);
							edgeIndicesToRelocate.add(index);
						} else if (isRightOf(childIdInTarget, nodeIdInTree, labeledTree) || childContainsDescendants){
							lastChild = Math.max(index, lastChild+1);
						} 
					}
					index++;
				}
				if (parentIndex != null){
					lastChild = parentIndex;
				}
				// add child at position of last child that will be removed or at the end
				// TODO: change this to use sibling information of the new node!
				Node newNodeToInsert = getNewNode(newNode, null, tree);
				parentBlock.addChildAt(newNodeToInsert, lastChild);
				int size = nodeToNewNodes.size();
				nodeToNewNodes.put(nodeIdInTree, newNodeToInsert);
				System.out.println("added "+newNodeToInsert.getName()+" to nodes. Size from " + size+" to " + nodeToNewNodes.size());
				for (Edge edgeToRelocate : edgesToRelocate){
					System.out.println("Relocating edge: "+edgesToRelocate+" to new ");
					Node target = edgeToRelocate.getTarget();
					parentBlock.removeOutgoingEdge(edgeToRelocate);
					target.removeIncomingEdge(edgeToRelocate);
					tree.removeEdge(edgeToRelocate);
					Block newBlock = (Block) newNodeToInsert;
					Edge childEdge = newBlock.addChild(target);
					target.addIncomingEdge(childEdge);
					if (target.getParents().size()>1){
						System.out.println("Debug me!");
					}	
				}
				if (newNode.getParents().size()>1){
					System.out.println("Debug me!");
				}
				if (newNodeToInsert.getParents().size()>1){
					System.out.println("Debug me!");
				}
				
			} else {
				System.out.println("Debug me!");
			}
			int[] indices = new int[edgeIndicesToRelocate.size()];
			int i = 0;
			for (Integer index : edgeIndicesToRelocate){
				indices[i++] = index;
			}
			oper.setChildIndices(indices);
		}
	}


	private boolean containsDescendants(Node child, Integer nodeIdInTree, LabeledTree labeledTree, BiMap<Integer,Node> nodeToNewNodes) {
		if (child instanceof Block){
			Block childBlock = (Block) child;
			boolean isDescendant = false;
			for (Node grandChild : childBlock.getChildren()){
				isDescendant = isDescendant || containsDescendants(grandChild, nodeIdInTree, labeledTree, nodeToNewNodes);
			}
			return isDescendant;
		} else {
			Integer childId = nodeToNewNodes.inverse().get(child);
			if (childId!=null){
				return getAncestors(childId, labeledTree).contains(nodeIdInTree);
			}
		}
		return false;
	}

	private Node getNewNode(Node newNode, Node origNode, ProcessTree tree) {
		Node np = null;
		try {
			if(newNode instanceof AbstractTask.Manual){
				np = new AbstractTask.Manual(newNode.getID(), newNode.getName());
			} else if(newNode instanceof AbstractTask.Automatic){
				np = new AbstractTask.Automatic(newNode.getID(), newNode.getName());
			} else {
				np = newNode.getClass().getConstructor(UUID.class, String.class).newInstance(newNode.getID(), newNode.getName());
			}
		} catch (InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException
				| NoSuchMethodException | SecurityException e) {
			e.printStackTrace();
		}
		tree.addNode(np);
		return np;
	}

	/**
	 * Checks whether a node is right of a child in the tree
	 * 
	 * @param childIdInTarget node id 
	 * @param nodeIdInTarget
	 * @param tree
	 * @return
	 */
	private boolean isRightOf(Integer childIdInTarget, int nodeIdInTarget, LabeledTree tree) {
		
		int[] orderedIndices = new int[tree.size()];
		populateArray(tree.getRoot(), orderedIndices, 0, tree);
		System.out.println(Arrays.toString(orderedIndices));
		for (Integer i : orderedIndices){
			if (i == childIdInTarget){
				return true;
			}
			if (i == nodeIdInTarget){
				return false;
			}
		}
		throw new IllegalArgumentException("Neither child, nor node id in tree!?");
	}
//	private boolean isRightOf(Integer childIdInTarget, int nodeIdInTarget, LabeledTree tree) {
//		Collection<Integer> rightSiblings = getRightSiblings(childIdInTarget, tree);
//		int parentOfChild = tree.getParent(childIdInTarget);
//		while (LabeledTree.NOT_FOUND != parentOfChild){
//			rightSiblings.addAll(getRightSiblings(parentOfChild, tree));
//			parentOfChild = tree.getParent(parentOfChild);
//		}
//		return rightSiblings.contains(nodeIdInTarget);
//	}

	private Collection<Integer> getRightSiblings(Integer childIdInTarget, LabeledTree tree) {
		int node = childIdInTarget;
		int sibling = tree.getNextSibling(node);
		Collection<Integer> siblings = new ArrayList<>();
		while (sibling != -1){
			siblings.add(sibling);
			node = sibling;
			sibling = tree.getNextSibling(node);
		}
		return siblings;
	}

	public static int getPositionInParent(Node origNode, Node newParentNode) {
		if (newParentNode instanceof Block){
			Block parent = (Block) newParentNode;
			List<Node> children = parent.getChildren();
			for (int i = 0; i < children.size(); i++){
				Node child = children.get(i);
				if (origNode.equals(child)){
					return i;
				}
			}
		}
		System.out.println("Debug me! - Why is it not a child?");
		return 0;
	}

	/**
	 * Checks whether "child in target" is a descendant of "node in target".
	 * @param childId
	 * @param parentId
	 * @param tree
	 * @return
	 */
	private boolean isAncestor(Integer childId, int parentId, LabeledTree tree) {
		boolean isDescendant = false;
		Collection<Integer> ancestors = getAncestors(childId, tree);
		isDescendant = ancestors.contains(parentId);
		int child = tree.getFirstChild(childId); 
		while (child != LabeledTree.NOT_FOUND){
			isDescendant = isDescendant || isAncestor(child, parentId, tree);
			child = tree.getNextSibling(child);
		}
		return isDescendant;
	}

	private Collection<Integer> getAncestors(Integer childIdInTarget, LabeledTree tree) {
		if (childIdInTarget == null){
			System.out.println("debug this.");
		}
		int child = childIdInTarget;
		int parent = tree.getParent(childIdInTarget);
		Collection<Integer> parents = new ArrayList<>();
		while (parent != -1){
			parents.add(parent);
			child = parent;
			parent = tree.getParent(child);
		}
		return parents;
	}

	private int findNextMappedParent(int nodeIdInTarget, LabeledTree tree, BiMap<Integer, Node> fromToNewMapping) {
		boolean found = false;
		int nodeId = nodeIdInTarget;
		while (!found && tree.getParent(nodeId) >= -1){
			nodeId = tree.getParent(nodeId);
			if (fromToNewMapping.containsKey(nodeId)){
				found = true;
			}
		}
		if (!found){
			return -1;
		}
		return nodeId;
	}

	/**
	 * Converts a mapping into a list of edit operations {@link EditOperation}.
	 * Assumes that {@link #map} is already computed. 
	 * @return a list of edit operations that transform the original tree to the target tree.
	 */
	private List<EditOperation> getEditOperations() {
		if (this.origEditOperations!=null){
			return new ArrayList<>(this.origEditOperations);
		} else {
			List<EditOperation> editOps = new ArrayList<>();
	
			for (Integer deletedNode : map.getAllDeletion()){
				EditOperation deleteOp = new EditOperation(Op.DELETE, origNodeIds.inverse().get(deletedNode), names.get(origNodeIds.inverse().get(deletedNode)), null, null, score.delete(deletedNode));
				editOps.add(deleteOp);
			}
			
			for (Integer insertedNode : map.getAllInsertion()){
				EditOperation insertOp = new EditOperation(Op.INSERT, null, null, targetNodeIds.inverse().get(insertedNode), names.get(targetNodeIds.inverse().get(insertedNode)), score.insert(insertedNode));
				editOps.add(insertOp);
			}
			
			for (int[] replacement : map.getAllReplacement()){
				double cost = score.replace(replacement[0], replacement[1]);
				if (cost > 0){
					EditOperation replaceNode = new EditOperation(Op.RENAME, origNodeIds.inverse().get(replacement[0]), names.get(origNodeIds.inverse().get(replacement[0])), targetNodeIds.inverse().get(replacement[1]), names.get(targetNodeIds.inverse().get(replacement[1])), cost);
					editOps.add(replaceNode);
				} else {
	//				System.out.println("From: "+names.get(fromOrigToNewNodes.get(replacement[0]).getID())+" - "+names.get(targetNodeIds.inverse().get(replacement[1]))+" costs nothing!!");
					// nothing changed here
				}
			}
			if (TreeMoverConfig.getInstance().getHeuristic().equals(Heuristic.RANDOM)){
				Collections.shuffle(editOps);
			}
			
			this.origEditOperations = editOps;
			return editOps;
		}
	}	
	
	private void visualizeTree(ProcessTree... trees){
		graphPanel.removeAll();
		graphPanel.setLayout(new GridLayout(trees.length, 1));
		
		for (ProcessTree tree : trees){
			TreeLayoutBuilder builder = new TreeLayoutBuilder(tree);
			JGraph graph = builder.getJGraph();
			graphPanel.add(new JScrollPane(graph));
		}
		
		graphPanel.revalidate();
		graphPanel.repaint();
	}
	
	private int populateArray(int root, int[] positions, int idx, LabeledTree tree){
		for (int child = tree.getFirstChild(root); child != LabeledTree.NOT_FOUND; child = tree.getNextSibling(child)){
			idx = populateArray(child, positions, idx, tree);
		}
		positions[idx] = root;
		return idx+1;
	}
	/**
	 * Computes the balanced distance that we want to cover in the first step.
	 * It respects the distance that we want to bridge by moving in the log later on.
	 * 
	 * @param trustLevelFirst
	 * @param trustLevelSecond
	 * 
	 * @return
	 */
	public static double getTrustLevelDistance(double trustLevelFirst, double trustLevelSecond) {
		if (trustLevelSecond+trustLevelFirst==2){
			return 0;
		}
		double maxTotalRepair = 2 - trustLevelSecond - trustLevelFirst;
		double actualTotalRepair = Math.min(1, maxTotalRepair);
		return ((1-trustLevelFirst) / maxTotalRepair) * actualTotalRepair;
	}
	/**
	 * Computes the remaining ratio that we need to cover to ensure the correct ratios of repaired inputs based on the trusts.
	 * @param trustFirst
	 * @param trustSecond
	 * @return
	 */
	public static double getTrustLevelSecondStepDistance(double trustFirst, double trustSecond) {
		double repairSecondRatio = TreeMover.getTrustLevelDistance(trustSecond, trustFirst);
		double repairFirstRatio = TreeMover.getTrustLevelDistance(trustFirst, trustSecond);
		double noRepairRatio = 1 - repairFirstRatio - repairSecondRatio;
		return  repairSecondRatio / (repairSecondRatio + noRepairRatio);
	}
}
