package org.processmining.generalizedconformance.experiment;

import lpsolve.LpSolve;
import nl.tue.astar.AStarThread.Canceller;
import org.apache.commons.io.FileUtils;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.in.XUniversalParser;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.jbpt.bp.RelSet;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.generalizedconformance.Utils;
import org.processmining.generalizedconformance.algorithms.converter.RelaxedPT2PetrinetConverter;
import org.processmining.generalizedconformance.algorithms.heuristic.bp.BehaviourJaccardSimilarity;
import org.processmining.generalizedconformance.algorithms.heuristic.footprint.ProcessTreeFootprintCreator;
import org.processmining.generalizedconformance.algorithms.heuristic.processtree.NodeWrapper;
import org.processmining.generalizedconformance.algorithms.miner.GeneralizedMiner;
import org.processmining.generalizedconformance.algorithms.miner.QualityCriterion;
import org.processmining.generalizedconformance.algorithms.mover.TreeMoverConfig;
import org.processmining.generalizedconformance.algorithms.mover.TreeMoverConfig.Heuristic;
import org.processmining.generalizedconformance.filter.SimpleHeuristicFilter;
import org.processmining.generalizedconformance.plugins.GeneralizedMinerPlugin;
import org.processmining.modelrepair.parameters.RepairConfiguration;
import org.processmining.modelrepair.plugins.Uma_RepairModel_Plugin;
import org.processmining.models.connections.petrinets.EvClassLogPetrinetConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.jgraph.ProMJGraphVisualizer;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.metrics.FitnessReplay;
import org.processmining.plugins.etm.fitness.metrics.Generalization;
import org.processmining.plugins.etm.fitness.metrics.PrecisionEscEdgesImproved;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.plugins.stochasticpetrinet.StochasticNetUtils;
import org.processmining.processtree.ProcessTree;
import org.processmining.ptconversions.pn.ProcessTree2Petrinet.PetrinetWithMarkings;
import org.processmining.xeslite.lite.factory.XFactoryLiteImpl;
import org.uncommonseditedbyjoosbuijs.watchmaker.framework.FitnessEvaluator;

import javax.swing.*;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import java.awt.*;
import java.io.File;
import java.util.*;
import java.util.List;

/**
 * Experiment for the BPM2016 submission "In Log and Model We Trust?"
 * 
 * compares proposed generalized conformance checking approach with:
 * http://dx.doi.org/10.1016/j.is.2013.12.007 
 * (Dirk Fahland, Wil M.P. van der Aalst: Model repair - aligning process models to reality)
 * 
 * 
 * 
 *  
 * @author Andreas R-Solti
 *
 */
public class BPM2016 {
	private static final String SEP = ";";
	private PluginContext context;
	private XLog inputLog;
	private String logName;
	
	private boolean debug = false;
	private JComponent debugScreen;
	
	private static final String[] FILES = new String[]{"bpi_2012_only_A.xes", "output_SBWL.xes", "dfci.xes" };
	

	private BPM2016(File logFile, String name) throws Exception {
		init(logFile, name);
	}
	
	private void init(File logFile, String name) throws Exception {
		this.context = StochasticNetUtils.getDummyConsoleProgressContext(); // TODO: perhaps set up a mock plugin
		this.logName = name;
		// the L in the paper
		this.inputLog = Utils.flattenLifecycles(loadLog(logFile));
		XConceptExtension.instance().assignName(this.inputLog, name);
		
		if (debug){
			JFrame frame = new JFrame("Debug Screen");
			this.debugScreen = new JPanel(new BorderLayout());
			this.debugScreen.setPreferredSize(new Dimension(1000, 800));
			frame.getContentPane().add(this.debugScreen, BorderLayout.CENTER);
			frame.pack();
			frame.setVisible(true);
		}
	}

	public static void main(String[] args) throws Exception {
		System.loadLibrary("lpsolve55");
		System.loadLibrary("lpsolve55j");
		System.out.println(LpSolve.lpSolveVersion());
		
		XFactoryLiteImpl.register();
		for (XFactory factory : XFactoryRegistry.instance().getAvailable()){
			if (factory instanceof XFactoryLiteImpl){
				XFactoryRegistry.instance().setCurrentDefault(factory);
			}
		}
		
	
		
		// the file to store the test log
		for (String fileName : FILES){
			for (Heuristic heuristic : Heuristic.values()){
				TreeMoverConfig.getInstance().setHeuristic(heuristic);
				runMe(fileName);		
			}
		}
		
	}

	protected static void runMe(String fileName) throws Exception {
		String prefix = "./tests/testfiles/bpm2016/";
		
//		String testLogFile = "./tests/testfiles/bpm2016/dfci.xes"; String name = "dfci3";
//		String testLogFile = "./tests/testfiles/bpm2016/BPI_Challenge_2013_incidents.xes"; String name = "bpi_2013";
//		String testLogFile = "./tests/testfiles/bpm2016/bpi_2011.xes";
//		String testLogFile = "./tests/testfiles/bpm2016/loan_process_filtered_only_A.xes"; String name = "loan_only_A";
//		String testLogFile = "./tests/testfiles/bpm2016/output_SBWL.xes"; String name = "sbwl_old_heuristic";
//		String testLogFile = "./tests/testfiles/bpm2016/bpi_2012.xes.gz"; String name = "bpi_2012";
		
		String name = fileName.substring(0, fileName.length()-4);
		
		String testLogFile = prefix + fileName;
		
		name += TreeMoverConfig.getInstance().getHeuristic().name();
		
		File logFile = new File(testLogFile);
		if (!logFile.exists()){
			throw new RuntimeException("Please download the BPI challenge 2013 log from \n"
					+ "http://data.3tu.nl/repository/uuid:500573e6-accc-4b0c-9576-aa5468b10cee\n"
					+ "and place it in the "+logFile.getParent()+" folder. Then you can re-run this experiment.");
		}
		
		BPM2016 experiment = new BPM2016(logFile, name);
		experiment.run();
	}
	
	
	private void run() throws Exception {
		// the M in the paper
//		HighFrequencyFilterPlugin filterPlugin = new HighFrequencyFilterPlugin();
//		HighFrequencyFilterParameters filterParams = new HighFrequencyFilterParameters(inputLog);
//		filterParams.setFrequencyThreshold(10);
//		filterParams.setDistanceThreshold(1);
//		XLog filteredInput = filterPlugin.run(context, inputLog, filterParams);
		
		XLog filteredInput = SimpleHeuristicFilter.filterLogForMostFrequentTrace(inputLog);
		
		System.out.println("Before filtering: "+inputLog.size()+", after filtering: "+filteredInput.size());
 		ProcessTree inputTree = Utils.mineProcessTree(filteredInput, 0.8);
// 		// remove also a child:
// 		Iterator<Node> nodeIter = inputTree.getNodes().iterator();
// 		Node node = nodeIter.next();
// 		while (!node.isLeaf() && nodeIter.hasNext()){
// 			node = nodeIter.next();
// 		}
// 		org.processmining.processtree.Edge edge = node.getIncomingEdges().iterator().next();
// 		node.removeIncomingEdge(edge);
// 		edge.getSource().removeOutgoingEdge(edge);
// 		inputTree.removeNode(node);
// 		inputTree.removeEdge(edge);
		
		// Petri net version of M
		PetrinetWithMarkings inputPetrinet = RelaxedPT2PetrinetConverter.convert(inputTree);
		
		
		// repair with different trust levels:
		double[] trustLevels = new double[]{0.,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.};
//		double trustInModel = 0.5;
		double[] trustInLogBounds = new double[]{0.0,1};
		int logResolution = 10;
//		double[] trustInLog = new double[]{1};
		
		GeneralizedMiner miner = GeneralizedMinerPlugin.mineGeneralLogModel(context, inputLog, inputTree);
		
		ProcessTree lOfM = Utils.mineProcessTree(inputLog, 0.0);
		JFrame newFrame = new JFrame("Visual comparison of inputs");
		newFrame.getContentPane().setLayout(new GridLayout(2, 1));
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(new TitledBorder(new EtchedBorder(), "input Model M"));
		panel.add(Utils.getGraphForTree(inputTree), BorderLayout.CENTER);
		JPanel panel2 = new JPanel(new BorderLayout());
		panel2.setBorder(new TitledBorder(new EtchedBorder(), "mined M(L)"));
		panel2.add(Utils.getGraphForTree(lOfM), BorderLayout.CENTER);
		
		newFrame.getContentPane().add(panel);
		newFrame.getContentPane().add(panel2);
		newFrame.pack();
		newFrame.setVisible(true);
		
		// evaluate quality criteria on models:
		StringBuffer buffer = new StringBuffer();
		buffer.append(getHeader()).append("\n");
		
		File resultFile = new File("./BPM16_results_"+logName+".csv");
		if (!resultFile.exists()) resultFile.createNewFile();
		FileUtils.write(resultFile, buffer.toString());
		
		// our approach:
		for (double trustInModel : trustLevels){
			for (int i = 0; i<=logResolution; i++){
				double trustInLog = trustInLogBounds[0]+(i/(double)logResolution)*(trustInLogBounds[1]-trustInLogBounds[0]);
				Pair<XLog, ProcessTree> pair = miner.getFittingPair(trustInLog, trustInModel);
				// ProcessTree version of M*
				ProcessTree bestTree = pair.getSecond();
				
				if (trustInModel==0. && trustInLog==1.){
					if (!miner.getTargetTree().toString().equals(bestTree.toString())){
						System.err.println("Debug me!!");
					}
				}
				// Petri net version of M*
				PetrinetWithMarkings bestPetrinet = RelaxedPT2PetrinetConverter.convert(bestTree);
//				context.addConnection(new InitialMarkingConnection(bestPetrinet.petrinet, bestPetrinet.initialMarking));
//				context.addConnection(new FinalMarkingConnection(bestPetrinet.petrinet, bestPetrinet.finalMarking));
//				String pnmlString = PnmlExportNet.exportPetriNetToPNMLOrEPNMLString(context, bestPetrinet.petrinet, PnmlType.PNML, true);
//				FileUtils.write(new File("./tests/testfiles/bpm2016/models/"+trustInModel+"_"+logName+".pnml"), pnmlString);
				
				buffer.append(evaluateQualityCriteria(inputLog, "M*_"+trustInModel, pair.getFirst(), bestPetrinet, inputPetrinet, bestTree, inputTree, trustInModel, trustInLog, this.debugScreen));
				FileUtils.write(resultFile, buffer.toString());		
			}
		}
		
		// repair M with "model repair approach"
		RepairConfiguration config = new RepairConfiguration(); // use default configuration
		
		Uma_RepairModel_Plugin modelRepairer = new Uma_RepairModel_Plugin();
		XLogInfo info = XLogInfoFactory.createLogInfo(inputLog, new XEventNameClassifier());
		TransEvClassMapping transEvClassMapping = StochasticNetUtils.getEvClassMapping(inputPetrinet.petrinet, inputLog);
		EvClassLogPetrinetConnection connection = new EvClassLogPetrinetConnection("connection", inputPetrinet.petrinet, inputLog,info.getEventClassifiers().iterator().next() , transEvClassMapping);
		context.getConnectionManager().addConnection(connection);
		Object[] netAndMarking = modelRepairer.repairModel_getT2Econnection(context, inputLog, inputPetrinet.petrinet, inputPetrinet.initialMarking, inputPetrinet.finalMarking, config);
		PetrinetWithMarkings repairedModel = new PetrinetWithMarkings();
		repairedModel.petrinet = (Petrinet) netAndMarking[0];
		repairedModel.initialMarking = (Marking) netAndMarking[1];
		repairedModel.finalMarking = StochasticNetUtils.getFinalMarking(context, repairedModel.petrinet);
		
		buffer.append(evaluateQualityCriteria(inputLog, "repaired Model",inputLog, repairedModel, inputPetrinet, null, null, Double.NaN, Double.NaN));
		
		
		
		// Joos' approach?
		System.out.println(buffer.toString());
		FileUtils.write(resultFile, buffer.toString());
		
	}
	
	public static String evaluateQualityCriteria(XLog inputLog, String modelName, XLog resultLog, PetrinetWithMarkings petrinet,
			 PetrinetWithMarkings inputPetrinet, ProcessTree bestTree, ProcessTree inputTree, Double trustInModel, Double trustInLog) {
		return evaluateQualityCriteria(inputLog, modelName, resultLog, petrinet, inputPetrinet, bestTree, inputTree, trustInModel, trustInLog, null);
	}
	/**
	 * 
	 * @param inputLog {@link XLog} the L in the paper
	 * @param modelName
	 * @param resultLog 
	 * @param petrinet
	 * @param inputPetrinet 
	 * @param bestTree
	 * @param inputTree {@link ProcessTree} the M in the paper 
	 * @param trustInModel
	 * @param trustInLog
	 * @return
	 */
 	public static String evaluateQualityCriteria(XLog inputLog, String modelName, XLog resultLog, PetrinetWithMarkings petrinet,
			 PetrinetWithMarkings inputPetrinet, ProcessTree bestTree, ProcessTree inputTree, Double trustInModel, Double trustInLog, JComponent debugScreen) {
		// replay both models on the log.
		StringBuffer result = new StringBuffer();
		
		if (debugScreen!=null){
			// show the petri net!
			ProMJGraphPanel newGraphPanel = ProMJGraphVisualizer.instance().visualizeGraphWithoutRememberingLayout(petrinet.petrinet);
			debugScreen.removeAll();
			debugScreen.add(newGraphPanel, BorderLayout.CENTER);
			debugScreen.revalidate();
			debugScreen.repaint();
		}
		
		
		
		result.append(XConceptExtension.instance().extractName(inputLog)).append(SEP); // logname;
		result.append(modelName).append(SEP); // logname;
		result.append(trustInModel).append(SEP).append(trustInLog).append(SEP); // trustM;trustL;
		
		Map<QualityCriterion, Object> qualityRepaired = Utils.getDistance(petrinet, resultLog);
		
		if(bestTree!=null){
			qualityRepaired.put(QualityCriterion.TREE_TRACE, bestTree.toString());
			RelaxedPT2PetrinetConverter.prepareForQualityCheck(bestTree);
			
			ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree();
			NAryTree convertedTree = converter.convert(bestTree);
			List<NAryTree> population = new ArrayList<>();
			population.add(convertedTree);
			CentralRegistry reg = new CentralRegistry(resultLog, new Random());
			FitnessEvaluator<NAryTree> evaluator = new FitnessReplay(reg,  new Canceller() {
				public boolean isCancelled() {
					return false;
				}
			});
			
			double fitness = evaluator.getFitness(convertedTree, population);
			FitnessEvaluator<NAryTree> evalPrecision = new PrecisionEscEdgesImproved(reg);
			double precision = evalPrecision.getFitness(convertedTree, population);
			FitnessEvaluator<NAryTree> evalGeneraliziation = new Generalization(reg);
			double generalization = evalGeneraliziation.getFitness(convertedTree, population);
			
			qualityRepaired.put(QualityCriterion.FITNESS_TREE, fitness);
			qualityRepaired.put(QualityCriterion.PRECISION_TREE, precision);
			qualityRepaired.put(QualityCriterion.GENERALIZATION_TREE, generalization);
			qualityRepaired.put(QualityCriterion.SIMPLICITY_TREE, (double) bestTree.getNodes().size());
			
			qualityRepaired.put(QualityCriterion.LOG_SIMILARITY, Utils.getTraceSimilarity(inputLog, resultLog));
			qualityRepaired.put(QualityCriterion.MODEL_SIMILARITY, Utils.getModelSimilarity(inputTree, bestTree));
			
			RelSet<ProcessTree, NodeWrapper> inputProfile = ProcessTreeFootprintCreator.getInstance().deriveFootprint(inputTree);
			RelSet<ProcessTree, NodeWrapper> resultProfile = ProcessTreeFootprintCreator.getInstance().deriveFootprint(bestTree);
			ProcessTree targetTree = Utils.mineProcessTree(inputLog, 0.0);
			RelSet<ProcessTree, NodeWrapper> targetProfile = ProcessTreeFootprintCreator.getInstance().deriveFootprint(targetTree);
			BehaviourJaccardSimilarity<ProcessTree, NodeWrapper> sim = new BehaviourJaccardSimilarity<>(inputProfile, resultProfile);
			double simM_Mstar = sim.getSimilarity();
			sim = new BehaviourJaccardSimilarity<>(inputProfile, targetProfile);
			double simM_MofL = sim.getSimilarity();
			sim = new BehaviourJaccardSimilarity<>(resultProfile, targetProfile);
			double simMstar_MofL = sim.getSimilarity();
			
//			double simM_Mstar = NameBasedSimilarity.getNameBasedSimilarity(inputProfile, resultProfile);
//			double simM_MofL = NameBasedSimilarity.getNameBasedSimilarity(inputProfile, targetProfile);
//			double simMstar_MofL = NameBasedSimilarity.getNameBasedSimilarity(resultProfile, targetProfile);
			// due to triangle inequality, we know that simM_Mstar + simMstar_LofM >= simM_LofM
			if (simM_Mstar + simMstar_MofL < simM_MofL){
				System.err.println("Debug me! Triangle inequality violated!");
			} else {
				qualityRepaired.put(QualityCriterion.SIM_M_MSTAR, simM_Mstar);
				qualityRepaired.put(QualityCriterion.SIM_MSTAR_MOFL, simMstar_MofL);
				qualityRepaired.put(QualityCriterion.SIM_M_MOFL, simM_MofL);
			}
			
			
		} else if (inputTree != null){
			// Dirks' repair:
			// distance between models is #oldNodes / #newNodes, thus similarity is (1-distance) 
			qualityRepaired.put(QualityCriterion.LOG_SIMILARITY, 1.0);
			double modelSimilarity = 1- (Math.abs(petrinet.petrinet.getNodes().size()-inputPetrinet.petrinet.getNodes().size()) / petrinet.petrinet.getNodes().size());
			qualityRepaired.put(QualityCriterion.MODEL_SIMILARITY, modelSimilarity);
		}
		addToResult(result, qualityRepaired);
		result.append("\n"); 
		return result.toString();
	}

 	
	public static void addToResult(StringBuffer result, Map<QualityCriterion, Object> qualityRepaired) {
		result.append(qualityRepaired.get(QualityCriterion.FITNESS)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.PRECISION1)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.GENERALIZATION1)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.PRECISION)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.GENERALIZATION)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.SIMPLICITY)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.FITNESS_TREE)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.PRECISION_TREE)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.GENERALIZATION_TREE)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.SIMPLICITY_TREE)).append(SEP);
		
		result.append(qualityRepaired.get(QualityCriterion.LOG_SIMILARITY)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.MODEL_SIMILARITY)).append(SEP);
		
		result.append(qualityRepaired.get(QualityCriterion.SIM_M_MOFL)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.SIM_M_MSTAR)).append(SEP);
		result.append(qualityRepaired.get(QualityCriterion.SIM_MSTAR_MOFL)).append(SEP);
		
		result.append(qualityRepaired.get(QualityCriterion.TREE_TRACE));
		
	}

	public static String getHeader() {
		return "logname"+SEP+"approach"+SEP+"trustM"+SEP+"trustL"+SEP+
				"fitness"+SEP+"precision"+SEP+"generalization"+SEP+"precision2"+SEP+"generalization2"+SEP+"simplicity"+SEP+
				"fitness_tree"+SEP+"precision_tree"+SEP+"generalization_tree"+SEP+"simplicity_tree"+SEP+
				"logSim"+SEP+"modelSim"+SEP+
				"simM_MofL"+SEP+"simM_Mstar"+SEP+"simMstar_MofL"+SEP+
				"tree";
	}

	/**
	 * Opens a Log from a given file.
	 * 
	 * @param file {@link File} containing the log.
	 * @return the loaded {@link XLog} 
	 * @throws Exception
	 */
	public static XLog loadLog(File file) throws Exception {
		XUniversalParser parser = new XUniversalParser();
		Collection<XLog> logs = parser.parse(file);
		if (logs.size() > 0){
			return logs.iterator().next();
		}
		return null;
	}
}
