package org.processmining.generalizedconformance.experiment;

import lpsolve.LpSolve;
import org.apache.commons.io.FileUtils;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.in.XUniversalParser;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.generalizedconformance.Utils;
import org.processmining.generalizedconformance.algorithms.converter.ProcessTreeConverter;
import org.processmining.generalizedconformance.algorithms.converter.RelaxedPT2PetrinetConverter;
import org.processmining.generalizedconformance.algorithms.miner.GeneralizedMiner;
import org.processmining.generalizedconformance.algorithms.mover.SimpleScore;
import org.processmining.generalizedconformance.algorithms.mover.TreeMoverConfig;
import org.processmining.generalizedconformance.algorithms.mover.TreeMoverConfig.Heuristic;
import org.processmining.generalizedconformance.plugins.GeneralizedMinerPlugin;
import org.processmining.log.parameters.HighFrequencyFilterParameters;
import org.processmining.log.plugins.HighFrequencyFilterPlugin;
import org.processmining.modelrepair.parameters.RepairConfiguration;
import org.processmining.modelrepair.plugins.Uma_RepairModel_Plugin;
import org.processmining.models.connections.petrinets.EvClassLogPetrinetConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.Semantics;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.petrinets.analysis.gedsim.utils.StringEditDistance;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.stochasticpetrinet.StochasticNetUtils;
import org.processmining.plugins.stochasticpetrinet.generator.Generator;
import org.processmining.plugins.stochasticpetrinet.generator.GeneratorConfig;
import org.processmining.plugins.stochasticpetrinet.simulator.PNSimulator;
import org.processmining.plugins.stochasticpetrinet.simulator.PNSimulatorConfig;
import org.processmining.processtree.ProcessTree;
import org.processmining.ptconversions.pn.ProcessTree2Petrinet.PetrinetWithMarkings;
import org.processmining.xeslite.lite.factory.XFactoryLiteImpl;
import treedist.EditScore;
import treedist.LabeledTree;
import treedist.Mapping;
import treedist.TreeEditDistance;

import javax.swing.*;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import java.awt.*;
import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * Experiment for the BPM2016 submission "In Log and Model We Trust?"
 * 
 * compares proposed generalized conformance checking approach with:
 * http://dx.doi.org/10.1016/j.is.2013.12.007 
 * (Dirk Fahland, Wil M.P. van der Aalst: Model repair - aligning process models to reality)
 * 
 * 
 * 
 *  
 * @author Andreas R-Solti
 *
 */
public class BPM2016Synthetic {
	private static final String SEP = ";";
	private static final String[] nameAlphabet = new String[]{"a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","v","x","y","z","A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z","0","1","2","3","4","5","6","7","8","9"};
	private PluginContext context;
	private XLog inputLog;
	private String logName;
	private ProcessTree inputTree;

	private BPM2016Synthetic () throws Exception {
		init("synthetic3");
	}
	
	private void init(String name) throws Exception {
		this.context = StochasticNetUtils.getDummyConsoleProgressContext(); // TODO: perhaps set up a mock plugin
		this.logName = name;
		// the L in the paper
		
		// generate two random models of size 20.
		GeneratorConfig config = new GeneratorConfig();
		config.setContainsLoops(true);
		config.setDegreeOfExclusiveChoices(10);
		config.setDegreeOfSequences(60);
		config.setDegreeOfParallelism(10);
		config.setDegreeOfLoops(10);
		
		config.setTransitionSize(15);
		
		Generator generator = new Generator(0);
		Object[] netAndMarkings = generator.generateStochasticNet(config);
		Petrinet net1 = (Petrinet) netAndMarkings[0];
		PNSimulator simulator = new PNSimulator();
		PNSimulatorConfig simConfig = new PNSimulatorConfig(1000);
		Semantics<Marking, Transition> semantics = StochasticNetUtils.getSemantics(net1);
		XLog log1 = simulator.simulate(null, net1, semantics, simConfig, (Marking) netAndMarkings[1]);
		
		Object[] netAndMarkingsSecond = generator.generateStochasticNet(config);
		Petrinet net2 = (Petrinet) netAndMarkingsSecond[0];
		semantics = StochasticNetUtils.getSemantics(net2);
		XLog log2 = simulator.simulate(null, net2, semantics, simConfig, (Marking) netAndMarkingsSecond[1]);
		
		// Mine corresponding trees:
		ProcessTree tree1 = Utils.mineProcessTree(log1, 0.1);
		ProcessTree tree2 = Utils.mineProcessTree(log2, 0.1);
		
		
		this.inputLog = log1;
		XConceptExtension.instance().assignName(this.inputLog, name);
		this.inputTree = tree2;
	}

	public static void main(String[] args) throws Exception {
		System.loadLibrary("lpsolve55");
		System.loadLibrary("lpsolve55j");
		System.out.println(LpSolve.lpSolveVersion());
		// the file to store the test log
		
		XFactoryLiteImpl.register();
		for (XFactory factory : XFactoryRegistry.instance().getAvailable()){
			if (factory instanceof XFactoryLiteImpl){
				XFactoryRegistry.instance().setCurrentDefault(factory);
			}
		}
			
		TreeMoverConfig.getInstance().setHeuristic(Heuristic.FOOTPRINT);
		
		BPM2016Synthetic experiment = new BPM2016Synthetic();
		experiment.run();
	}
	
	
	private void run() throws Exception {
		if (inputTree==null){
			// the M in the paper
			HighFrequencyFilterPlugin filterPlugin = new HighFrequencyFilterPlugin();
			HighFrequencyFilterParameters filterParams = new HighFrequencyFilterParameters(inputLog);
			filterParams.setFrequencyThreshold(10);
			filterParams.setDistanceThreshold(1);
			XLog filteredInput = filterPlugin.run(context, inputLog, filterParams);
			System.out.println("Before filtering: "+inputLog.size()+", after filtering: "+filteredInput.size());
	 		inputTree = Utils.mineProcessTree(filteredInput, 0.8);
		}
// 		// remove also a child:
// 		Iterator<Node> nodeIter = inputTree.getNodes().iterator();
// 		Node node = nodeIter.next();
// 		while (!node.isLeaf() && nodeIter.hasNext()){
// 			node = nodeIter.next();
// 		}
// 		org.processmining.processtree.Edge edge = node.getIncomingEdges().iterator().next();
// 		node.removeIncomingEdge(edge);
// 		edge.getSource().removeOutgoingEdge(edge);
// 		inputTree.removeNode(node);
// 		inputTree.removeEdge(edge);
		
		// Petri net version of M
		PetrinetWithMarkings inputPetrinet = RelaxedPT2PetrinetConverter.convert(inputTree);
		
		
		// repair with different trust levels:
		double[] trustLevels = new double[]{0.,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.};
//		double trustInModel = 0.5;
		double[] trustInLogBounds = new double[]{0.0,1};
		int logResolution = 10;
//		double[] trustInLog = new double[]{1};
		
		GeneralizedMiner miner = GeneralizedMinerPlugin.mineGeneralLogModel(context, inputLog, inputTree);
		
		ProcessTree lOfM = Utils.mineProcessTree(inputLog, 0.0);
		JFrame newFrame = new JFrame("Visual comparison of inputs");
		newFrame.getContentPane().setLayout(new GridLayout(2, 1));
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(new TitledBorder(new EtchedBorder(), "input Model M"));
		panel.add(Utils.getGraphForTree(inputTree), BorderLayout.CENTER);
		JPanel panel2 = new JPanel(new BorderLayout());
		panel2.setBorder(new TitledBorder(new EtchedBorder(), "mined M(L)"));
		panel2.add(Utils.getGraphForTree(lOfM), BorderLayout.CENTER);
		
		newFrame.getContentPane().add(panel);
		newFrame.getContentPane().add(panel2);
		newFrame.pack();
		newFrame.setVisible(true);
		
		// evaluate quality criteria on models:
		StringBuffer buffer = new StringBuffer();
		buffer.append(BPM2016.getHeader()).append("\n");
		
		File resultFile = new File("./BPM16_results_"+logName+".csv");
		if (!resultFile.exists()) resultFile.createNewFile();
		FileUtils.write(resultFile, buffer.toString());
		
		// our approach:
		for (double trustInModel : trustLevels){
			for (int i = 0; i<=logResolution; i++){
				double trustInLog = trustInLogBounds[0]+(i/(double)logResolution)*(trustInLogBounds[1]-trustInLogBounds[0]);
				Pair<XLog, ProcessTree> pair = miner.getFittingPair(trustInLog, trustInModel);
				// ProcessTree version of M*
				ProcessTree bestTree = pair.getSecond();
				// Petri net version of M*
				PetrinetWithMarkings bestPetrinet = RelaxedPT2PetrinetConverter.convert(bestTree);
//				context.addConnection(new InitialMarkingConnection(bestPetrinet.petrinet, bestPetrinet.initialMarking));
//				context.addConnection(new FinalMarkingConnection(bestPetrinet.petrinet, bestPetrinet.finalMarking));
//				String pnmlString = PnmlExportNet.exportPetriNetToPNMLOrEPNMLString(context, bestPetrinet.petrinet, PnmlType.PNML, true);
//				FileUtils.write(new File("./tests/testfiles/bpm2016/models/"+trustInModel+"_"+logName+".pnml"), pnmlString);
				
				buffer.append(BPM2016.evaluateQualityCriteria(inputLog, "M*_"+trustInModel, pair.getFirst(), bestPetrinet, inputPetrinet, bestTree, inputTree, trustInModel, trustInLog));
				FileUtils.write(resultFile, buffer.toString());		
			}
		}
		
		// repair M with "model repair approach"
		RepairConfiguration config = new RepairConfiguration(); // use default configuration
		
		Uma_RepairModel_Plugin modelRepairer = new Uma_RepairModel_Plugin();
		XLogInfo info = XLogInfoFactory.createLogInfo(inputLog, new XEventNameClassifier());
		TransEvClassMapping transEvClassMapping = StochasticNetUtils.getEvClassMapping(inputPetrinet.petrinet, inputLog);
		EvClassLogPetrinetConnection connection = new EvClassLogPetrinetConnection("connection", inputPetrinet.petrinet, inputLog,info.getEventClassifiers().iterator().next() , transEvClassMapping);
		context.getConnectionManager().addConnection(connection);
		Object[] netAndMarking = modelRepairer.repairModel_getT2Econnection(context, inputLog, inputPetrinet.petrinet, inputPetrinet.initialMarking, inputPetrinet.finalMarking, config);
		PetrinetWithMarkings repairedModel = new PetrinetWithMarkings();
		repairedModel.petrinet = (Petrinet) netAndMarking[0];
		repairedModel.initialMarking = (Marking) netAndMarking[1];
		repairedModel.finalMarking = StochasticNetUtils.getFinalMarking(context, repairedModel.petrinet);
		
		buffer.append(BPM2016.evaluateQualityCriteria(inputLog, "repaired Model",inputLog, repairedModel, inputPetrinet, null, null, Double.NaN, Double.NaN));
		
		
		
		// Joos' approach?
		System.out.println(buffer.toString());
		FileUtils.write(resultFile, buffer.toString());
		
	}
	

 	
 	/**
 	 * Model similarity is defined as (tree edit distance / max tree edits (i.e. remove all, add all others) )  
 	 * @param inputTree {@link ProcessTree} M
 	 * @param bestTree {@link ProcessTree} M*
 	 * @param targetTree {@link ProcessTree} M(L)
 	 * @return
 	 */
	private static Double getModelSimilarity(ProcessTree inputTree, ProcessTree resultTree) {
		ProcessTreeConverter converter = new ProcessTreeConverter();
		LabeledTree origLblTree = converter.getLabeledTree(inputTree);
		LabeledTree resultLblTree = converter.getLabeledTree(resultTree);
		
		Mapping map = new Mapping(origLblTree, resultLblTree);
		EditScore score = new SimpleScore(origLblTree, resultLblTree);
		TreeEditDistance dist = new TreeEditDistance(score);
		double distance = dist.calc(origLblTree, resultLblTree, map);
		return 1. - (distance / (inputTree.getNodes().size() + resultTree.getNodes().size()));
	}

	private static Double getTraceSimilarity(XLog inputLog, XLog resultLog) {
		int numEventsDifferent = 0;
		int allEvents = 0;
		if (inputLog.size() != resultLog.size()){
			throw new IllegalArgumentException("Logs must have equal number of traces!");
		}
		Map<String,String> eventEncoding = new HashMap<>();
		for (int trIndex = 0 ; trIndex < inputLog.size(); trIndex++){
			XTrace trInput = inputLog.get(trIndex);
			XTrace trResult = resultLog.get(trIndex);
			String inputString = getTraceString(trInput, eventEncoding);
			String resultString = getTraceString(trResult, eventEncoding);
			allEvents += inputString.length();
			allEvents += resultString.length();
			numEventsDifferent += StringEditDistance.editDistance(inputString, resultString);
		}
		return 1 - (numEventsDifferent / (double)allEvents);
	}

	private static String getTraceString(XTrace trInput, Map<String, String> eventEncoding) {
		StringBuffer buf = new StringBuffer();
		for (XEvent e : trInput){
			String evName = XConceptExtension.instance().extractName(e);
			if (!eventEncoding.containsKey(evName)){
				eventEncoding.put(evName, nameAlphabet[eventEncoding.size()]);
			}
			buf.append(eventEncoding.get(evName));
		}
		return buf.toString();
	}

	

	/**
	 * Opens a Log from a given file.
	 * 
	 * @param file {@link File} containing the log.
	 * @return the loaded {@link XLog} 
	 * @throws Exception
	 */
	private XLog loadLog(File file) throws Exception {
		XUniversalParser parser = new XUniversalParser();
		Collection<XLog> logs = parser.parse(file);
		if (logs.size() > 0){
			return logs.iterator().next();
		}
		return null;
	}
}
