package org.processmining.generalizedconformance.plugins;

import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.generalizedconformance.precision.EntropyPrecisionRecall;
import org.processmining.generalizedconformance.precision.automata.PrecisionRecallComputer;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.processtree.ProcessTree;
import org.processmining.ptconversions.pn.ProcessTree2Petrinet;

public class EigenvalueBasedPrecisionPlugin {

    @Plugin(name = "Eigenvalue-based Precision (ProcessTree)", parameterLabels = { "Event Log",
            "Process Tree" }, returnLabels = { "Precision" }, returnTypes = { EntropyPrecisionRecall.class}, userAccessible = true,
            help = "Computes Precision between event log and model.")
    @UITopiaVariant(affiliation = "Vienna University of Economics and Business", author = "Andreas Solti", email = "solti@ai.wu.ac.at", uiLabel = UITopiaVariant.USEPLUGIN)
    public EntropyPrecisionRecall getPrecision(PluginContext context, XLog log, ProcessTree tree) {
        try {
            long startTime = System.currentTimeMillis();
            ProcessTree2Petrinet.PetrinetWithMarkings pnWithMarkings = ProcessTree2Petrinet.convert(tree,true);
            EntropyPrecisionRecall precision = getPrecision(context, log, new AcceptingPetriNetImpl(pnWithMarkings.petrinet, pnWithMarkings.initialMarking, pnWithMarkings.finalMarking));
            precision.setTotalTime(System.currentTimeMillis()-startTime);
            return precision;
        } catch (ProcessTree2Petrinet.NotYetImplementedException e) {
            e.printStackTrace();
        } catch (ProcessTree2Petrinet.InvalidProcessTreeException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Plugin(name = "Eigenvalue-based Precision (Petrinet)", parameterLabels = { "Event Log",
            "Petrinet" }, returnLabels = { "Precision" }, returnTypes = { EntropyPrecisionRecall.class}, userAccessible = true,
            help = "Computes Precision between event log and model.")
    @UITopiaVariant(affiliation = "Vienna University of Economics and Business", author = "Andreas Solti", email = "solti@ai.wu.ac.at", uiLabel = UITopiaVariant.USEPLUGIN)
    public static EntropyPrecisionRecall getPrecision(PluginContext context, XLog log, Petrinet net) {
        return getPrecision(context, log, new AcceptingPetriNetImpl(net));
    }

    @Plugin(name = "Eigenvalue-based Precision (AcceptingPetriNet)", parameterLabels = { "Event Log",
            "Accepting Petrinet" }, returnLabels = { "Precision" }, returnTypes = { EntropyPrecisionRecall.class}, userAccessible = true,
            help = "Computes Precision between event log and model.")
    @UITopiaVariant(affiliation = "Vienna University of Economics and Business", author = "Andreas Solti", email = "solti@ai.wu.ac.at", uiLabel = UITopiaVariant.USEPLUGIN)
    public static EntropyPrecisionRecall getPrecision(final PluginContext context, XLog log, AcceptingPetriNet net){
        long startTime = System.currentTimeMillis();
        ProMCanceller canceller = new ProMCanceller() {
            @Override
            public boolean isCancelled() {
                if (context != null && context.getProgress() != null){
                    return context.getProgress().isCancelled();
                } else {
                    return false;
                }
            }
        };
        EntropyPrecisionRecall precision = PrecisionRecallComputer.getPrecisionAndRecall(context, canceller, log, net);
        if (canceller.isCancelled()){
            return null;
        }
        precision.setTotalTime(System.currentTimeMillis()-startTime);
        return precision;
    }
}
