package org.processmining.generalizedconformance.precision.automata;

import dk.brics.automaton2.Automaton;
import dk.brics.automaton2.RunAutomaton;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.framework.packages.impl.CancelledException;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.generalizedconformance.Utils;
import org.processmining.generalizedconformance.algorithms.converter.AcceptingPetriNet2Automaton;
import org.processmining.generalizedconformance.precision.EntropyPrecisionRecall;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.impl.EfficientStochasticNetSemanticsImpl;
import org.processmining.plugins.InductiveMiner.Pair;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.manifestreplayer.PNManifestReplayerParameter;
import org.processmining.plugins.petrinet.manifestreplayer.transclassifier.DefTransClassifier;
import org.processmining.plugins.petrinet.manifestreplayer.transclassifier.TransClasses;
import org.processmining.plugins.petrinet.manifestreplayresult.Manifest;
import org.processmining.plugins.stochasticpetrinet.StochasticNetUtils;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;
import org.processmining.projectedrecallandprecision.helperclasses.EfficientLog;
import org.processmining.projectedrecallandprecision.helperclasses.ProjectPetriNetOntoActivities;

import java.util.HashSet;
import java.util.Set;

public class PrecisionRecallComputer {

    private static final Log logger = LogFactory.getLog(PrecisionRecallComputer.class);

    /**
     * Computes Precision and Recall for event log and an accepting petri net. Uses the default Name Classifier to
     * establish a link between model and log.
     *
     * @param context {@link PluginContext} that can be null in testing or UI-less computation
     * @param canceller {@link ProMCanceller} a handler that indicates, whether computation should be aborted due to user cancellation
     * @param log {@link XLog} the event log to check for precision & fitness
     * @param net {@link AcceptingPetriNet} that has corresponding initial and final markings set
     * @return
     */
    public static EntropyPrecisionRecall getPrecisionAndRecall(PluginContext context, ProMCanceller canceller, XLog log, AcceptingPetriNet net) {
        try {
            return getPrecisionAndRecall(context, canceller, log, net, XLogInfoImpl.NAME_CLASSIFIER, null);
        } catch (CancelledException e) {
            logger.info("Precision computation cancelled!");
            return null;
        }
    }

    /**
     * Computes Precision and Recall for event log and an accepting petri net.
     * @param context {@link PluginContext} that can be null in testing or UI-less computation
     * @param canceller {@link ProMCanceller} a handler that indicates, whether computation should be aborted due to user cancellation
     * @param log {@link XLog} the event log to check for precision & fitness
     * @param net {@link AcceptingPetriNet} that has corresponding initial and final markings set
     * @param eventClassifier the classifier to be used (for example the name classifier XLogInfoImpl.NAME_CLASSIFIER)
     * @param resultL {@link EntropyResult} of a log (it is possible to pass in the result, to avoid recomputing it all the time.)
     * @return
     * @throws CancelledException
     */
    public static EntropyPrecisionRecall getPrecisionAndRecall(PluginContext context, ProMCanceller canceller, XLog log, AcceptingPetriNet net, XEventClassifier eventClassifier, EntropyResult resultL) throws CancelledException{
        String name = Utils.getName(net.getNet(),"M");
        String logName = Utils.getName(log,"L");

        if (context != null && context.getProgress() != null) {
            context.getProgress().setValue(0);
            context.getProgress().setMaximum(100);
        }
        log(context, "Starting precision computation for "+name, 1);

        EfficientStochasticNetSemanticsImpl semantics = new EfficientStochasticNetSemanticsImpl();
        Marking initMarking = net.getInitialMarking();
        semantics.initialize(net.getNet().getTransitions(), initMarking);

        // ensure that only fitting traces are in the log:
//        XLog fittingLog = XFactoryRegistry.instance().currentDefault().createLog(log.getAttributes());


//        filterForFittingTracesByAlignment(context, canceller, log, net, initMarking, eventClassifier, fittingLog);

//        log(context, "Aligned log to " + name + " and got " + fittingLog.size() + " fitting traces.", 25);
        checkCancelled(canceller);

//        if (fittingLog.isEmpty()){
//            log(context, "No fitting traces in log "+ logName+
//                    " that match model "+name+". Precision is 0 be definition!", 99);
//            return new EntropyPrecisionRecall(new EntropyResult(logName,0,0,0,0,0,0,0,0,0,0,0,true),
//                    new EntropyResult(name,0,0,0,0,0,0,0,2,1,0,0,true), 0);
//        }
//        double alignmentFittingRate = fittingLog.size() / (double)log.size();
        // prepare
        EfficientLog elog = new EfficientLog(log, eventClassifier);
        String[] activities = elog.getActivities();
        log(context, "Converted log to efficient log.", 28);
        if (canceller.isCancelled()){
            return null;
        }
        String[] names = getTransitionNames(net, activities);

        AcceptingPetriNet projectedNet = ProjectPetriNetOntoActivities.project(net, canceller, names);
        log(context, "Projected net to (all) activities.", 30);
        EntropyPrecisionRecall precision = null;
        Automaton a = AcceptingPetriNet2Automaton.convert(projectedNet, Integer.MAX_VALUE, canceller);
        log(context, "Converted net to automaton.", 35);
        checkCancelled(canceller);

        EntropyResult resultM = getResult(name, net.getNet().getNodes().size(), TopologicalEntropyComputer.getTopologicalEntropy(a, name, canceller));
        log(context, "Computed model automaton topology.", 60);
        checkCancelled(canceller);

        RunAutomaton ra = new RunAutomaton(a, canceller);
        double fittingTracesFraction;
        EntropyResult resultLM;
        {
            Pair<Double, Automaton> pair = processLog(elog, ra, true, canceller, names);
            fittingTracesFraction = pair.getA();
            log(context, "Projected log into model.", 70);
            checkCancelled(canceller);

            a = pair.getB(); // log automaton of accepted traces
            if (a == null) {
                log(context, "No fitting traces in log " + logName +
                        " that match model " + name + ". Precision is 0 be definition!", 79);
            }
            resultLM = getResult(Utils.getName(log, logName), elog.size(), TopologicalEntropyComputer.getTopologicalEntropy(a, logName, canceller));
            log(context, "Computed log-model automaton topology.", 75);
        }
        if (fittingTracesFraction >= 1){
            resultL = resultLM;
        }
        if (resultL == null) {
            resultL = getEntropyLogResult(context, canceller, log, logName, elog, names);
            checkCancelled(canceller);
        }
        return new EntropyPrecisionRecall(resultLM, resultM, resultL, fittingTracesFraction);
    }

    private static void checkCancelled(ProMCanceller canceller) throws CancelledException {
        if (canceller.isCancelled()){
            throw new CancelledException();
        }
    }

    public static EntropyResult getEntropyLogResult(PluginContext context, ProMCanceller canceller, XLog log, String logName, EfficientLog elog, String[] names) throws CancelledException{
        Automaton a;
        EntropyResult resultL;Pair<Double, Automaton> pair = processLog(elog, null, false, canceller, names);
        log(context, "Computed Log automaton.", 80);
        checkCancelled(canceller);

        a = pair.getB();
        resultL = getResult(Utils.getName(log, logName), elog.size(), TopologicalEntropyComputer.getTopologicalEntropy(a, logName, canceller));
        log(context, "Computed log automaton topological entropy.", 95);
        return resultL;
    }

    public static String[] getTransitionNames(AcceptingPetriNet net, String[] activities) {
        Set<String> transitionNames = new HashSet<>();
        for (org.processmining.models.graphbased.directed.petrinet.elements.Transition t : net.getNet().getTransitions()) {
            if (!t.isInvisible() && t.getLabel() != null) {
                transitionNames.add(t.getLabel());
            }
        }
        for (int i = 0; i<activities.length; i++) transitionNames.add(activities[i]);
        return transitionNames.toArray(new String[transitionNames.size()]);
    }

    private static void log(PluginContext context, String message, int degreeDone) {
        if (context != null && context.getProgress() != null) {
            context.getProgress().setValue(degreeDone);
            context.log(message);
        }
        logger.info(message);
    }

    private static EntropyResult getResult(String name, int size, EntropyResult entResult) {
        entResult.name = name;
        entResult.size = size;
        return entResult;
    }

    /**
     * Use automaton acceptance of each trace, as we are only interested in truly in the model (like in soccer).
     * To check language containment with an automaton is much faster than the alignment,
     * if we only are interested in completely fitting traces
     *
     * @param context
     * @param canceller
     * @param log
     * @param net
     * @param initMarking
     * @param eventClassifier
     * @param fittingLog
     */
    @Deprecated
    private static void filterForFittingTracesByAlignment(PluginContext context, ProMCanceller canceller, XLog log, AcceptingPetriNet net, Marking initMarking, XEventClassifier eventClassifier, XLog fittingLog) {
        TransEvClassMapping transitionEventClassMap = StochasticNetUtils.getEvClassMapping(net.getNet(), log, eventClassifier);
        TransClasses transClasses = new TransClasses(net.getNet(), new DefTransClassifier());
        PNManifestReplayerParameter parameters = StochasticNetUtils.getParameters(log, transitionEventClassMap, net.getNet(),
                net.getInitialMarking(), net.getFinalMarkings().iterator().next(),
                eventClassifier, transClasses);
        Manifest manifest = (Manifest) StochasticNetUtils.replayLog(context, net.getNet(), log, parameters, true);
        if (canceller.isCancelled()){
            return;
        }
        org.processmining.models.graphbased.directed.petrinet.elements.Transition[] idx2Trans = net.getNet().getTransitions().toArray(new org.processmining.models.graphbased.directed.petrinet.elements.Transition[net.getNet().getTransitions().size()]);

        for (int i = 0; i < manifest.getCasePointers().length; i++){
            int caseId = manifest.getCasePointers()[i];
            boolean fitting = true;
            if (caseId >= 0) {
                int[] man = manifest.getManifestForCase(i);
                int currIdx = 0;
                while (currIdx < man.length) {
                    if (man[currIdx] == Manifest.MOVELOG) {
                        fitting = false;
                        currIdx++;
                    } else if (man[currIdx] == Manifest.MOVEMODEL) {
                        if (!idx2Trans[man[currIdx + 1]].isInvisible()) {
                            fitting = false;
                        }
                        currIdx += 2;
                    } else if (man[currIdx] == Manifest.MOVESYNC) {
                        currIdx += 2;
                    }
                }
                if (fitting){
                    fittingLog.add(log.get(i));
                }
            }
        }
    }

    public static Pair<Double, Automaton> processLog(EfficientLog log, RunAutomaton modelAutomaton, boolean selectOnlyFittingTraces, ProMCanceller canceller, String... names) {
        short[] projectionKey = log.getProjectionKey(names);
        int replayableTraces = 0;
        Automaton logAutomaton = null;

        for(int traceIndex = 0; traceIndex < log.size(); ++traceIndex) {
            Pair<Boolean, Automaton> p = processTrace(modelAutomaton, logAutomaton, log, traceIndex, projectionKey, selectOnlyFittingTraces, canceller);
            if (canceller.isCancelled() || p == null) {
                return null;
            }

            logAutomaton = p.getB();
            if (p.getA()) {
                ++replayableTraces;
            }
        }
        if (logAutomaton != null) {
            logAutomaton.minimize(canceller);
        }
        if (canceller.isCancelled()) {
            return null;
        } else {
            return Pair.of(replayableTraces/(double)log.size(), logAutomaton);
        }
    }

    public static Pair<Boolean, Automaton> processTrace(RunAutomaton modelAutomaton, Automaton logAutomaton, EfficientLog log, int trace, short[] projectionKey, boolean onlyFittingTraces, ProMCanceller canceller) {
        short[] projectedTrace = log.getProjectedTrace(trace, projectionKey);
        boolean addToAutomaton = !onlyFittingTraces || modelAutomaton.run(projectedTrace);
        if (addToAutomaton) { // don't add trace to log automaton, if it can't be replayed.
            if (logAutomaton == null) {
                String automatonString = projectedTraceToString(projectedTrace);
                logAutomaton = Automaton.makeString(automatonString);
                logAutomaton.expandSingleton();
            } else {
                logAutomaton.incorporateTrace(projectedTrace, canceller);
                if (canceller.isCancelled()) {
                    return null;
                }
            }
        }

        return Pair.of(addToAutomaton, logAutomaton);
    }

    public static String projectedTraceToString(short[] projectedTrace) {
        String result = "";

        for(int event = 0; event < projectedTrace.length; ++event) {
            if (projectedTrace[event] >= 0) {
                result = result + (char)projectedTrace[event];
            }
        }

        return result;
    }
}
