package org.processmining.generalizedconformance.visualizer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.swing.*;
import java.awt.*;
import java.util.logging.Logger;

public class VennLabel extends JLabel {

    private static final Log log = LogFactory.getLog(VennLabel.class);

    public static final double RADIUS = 200;

    public static final Color LOG_COLOR = new Color(0,255,0,200);
    public static final Color MODEL_COLOR = new Color(255,255,0,120);

    private final double sizeModel;
    private final double sizeLog;
    private final double areaIntersection;

    private final Circle logCircle;
    private final Circle modelCircle;

    public VennLabel(double eigenLM, double eigenM, double eigenL, double percentFitting) {
        // size of model is the basis
        double tempSizeModel = 1;
        // fitting fraction, if it is not empty
        double sizeLogFitting = percentFitting > 0 ? eigenLM / eigenM : 0;
        // size of log is simply the fraction of the log vs. the model
        double tempSizeLog = (eigenL / eigenM) * tempSizeModel; // (1-percentFitting) * tempSizeModel + sizeLogFitting;
        if (eigenL<=0){
            tempSizeLog = 0.01*tempSizeModel;
        }


        if (tempSizeModel > tempSizeLog){
            sizeModel = tempSizeModel * RADIUS;
        } else {
            sizeModel = tempSizeModel / tempSizeLog * RADIUS;
        }

        sizeLog = Math.sqrt(tempSizeLog / tempSizeModel) * sizeModel;

        areaIntersection = sizeLogFitting * (Math.PI * sizeModel * sizeModel);

        logCircle = new Circle(RADIUS, RADIUS, sizeLog);

        int minX = (int) RADIUS;
        int maxX = (int) (RADIUS + sizeLog+sizeModel+1);
        int x = (int)((minX+maxX)/2.);
        double intersectArea = getIntersection(sizeLog, sizeModel, x-RADIUS);
        while (Math.abs(intersectArea - areaIntersection) > 5){
            if (intersectArea < areaIntersection){
                maxX = x-1;
            } else {
                minX = x+1;
            }
            if (minX >= maxX) {
                x = maxX;
                break;
            }
            x = (int)((minX+maxX)/2.);
            intersectArea = getIntersection(sizeLog, sizeModel, x-RADIUS);
        }
//        while (intersectArea > areaIntersection){
//            x += 1;
//            intersectArea = getIntersection(sizeLog, sizeModel, x-RADIUS);
//        }
        modelCircle = new Circle(x, RADIUS, sizeModel);
        double areaModel = sizeModel*sizeModel;
        double areaLog = sizeLog*sizeLog;
        log.info(String.format("Model vs Log size: %.2f",areaModel/areaLog));
    }

    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        // draw two circles:
        int alpha = 127; // 50% transparent
        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        int xOffset = (int)((getSize().width / 2)-RADIUS-(modelCircle.getCenterX() - logCircle.getCenterX())/2.);
        int yOffset = (int)((getSize().height / 2)-RADIUS);
        draw(g2, logCircle, LOG_COLOR,"L", true, xOffset, yOffset, false);
        draw(g2, modelCircle, MODEL_COLOR, "M", false, xOffset, yOffset, false);
        draw(g2, logCircle, LOG_COLOR, "L", true, xOffset, yOffset, true);
        draw(g2, modelCircle, MODEL_COLOR, "M", false, xOffset, yOffset, true);
    }

    private void draw(Graphics2D g2, Circle logCircle, Color color, String l, boolean left, int xOffset, int yOffset, boolean onlyLabel) {
        Color c = g2.getColor();
        if (!onlyLabel) {
            g2.setColor(color);
            g2.fillOval(xOffset + (int) (logCircle.getCenterX() - logCircle.getRadius()), yOffset + (int) (logCircle.getCenterY() - logCircle.getRadius()), (int) (logCircle.getRadius() * 2.), (int) (logCircle.getRadius() * 2.));
            g2.setColor(color.darker().darker());
            g2.setStroke(new BasicStroke(2f));
            g2.drawOval(xOffset + (int) (logCircle.getCenterX() - logCircle.getRadius()), yOffset + (int) (logCircle.getCenterY() - logCircle.getRadius()), (int) (logCircle.getRadius() * 2.), (int) (logCircle.getRadius() * 2.));
            g2.setColor(c);
        }
        g2.drawString(l,xOffset+(float)(logCircle.getCenterX()-(left?logCircle.getRadius()/2.:-logCircle.getRadius()/2.)), yOffset+(float)(logCircle.getCenterY()));
    }

    private static double getIntersection(double radius1, double radius2, double distance) {
        Double r = radius1;
        Double R = radius2;
        Double d = distance;
        if (radius1+radius2<=distance){
            return 0;
        }
        if(R < r){
            // swap
            r = radius2;
            R = radius1;
        }
        if (distance <= R - r){ // completely included!
            return r*r*Math.PI;
        }
        Double part1 = r*r*Math.acos((d*d + r*r - R*R)/(2*d*r));
        Double part2 = R*R*Math.acos((d*d + R*R - r*r)/(2*d*R));
        Double part3 = 0.5*Math.sqrt((-d+r+R)*(d+r-R)*(d-r+R)*(d+r+R));

        return part1 + part2 - part3;
    }
}
